# from __future__ import absolute_import
#
# from collections import OrderedDict
# from inspect import cleandoc
# from custom_inherit._doc_parse_tools.numpy_parse_tools import \
#         parse_numpy_doc
#
# __all__ = ["merge_numpy_docs"]
#
#
# def merge_section(key, prnt_sec, child_sec, merge_within_sections=True):
#     """ Synthesize a output numpy docstring section.
#     Parameters
#     ----------
#     key: str
#         The numpy-section being merged.
#     prnt_sec: Optional[str]
#         The docstring section from the parent's attribute.
#     child_sec: Optional[str]
#         The docstring section from the child's attribute.
#     merge_within_sections: bool
#         merge parents and child docstring
#     Returns
#     -------
#     Optional[str]
#         The output docstring section.
#     """
#
#     doc_sections_that_cant_merge = [
#         "Short Summary",
#         "Deprecation Warning",
#         "Extended Summary",
#         "Examples"
#     ]
#
#     def common_start(sa, sb):
#         """ returns the longest common substring from the beginning of sa and sb """
#         def _iter():
#             for a, b in zip(sa, sb):
#                 if a == b:
#                     yield a
#                 else:
#                     return
#
#         return ''.join(_iter())
#
#     if prnt_sec is None and child_sec is None:
#         return None
#
#     if key == "Short Summary":
#         header = ""
#     else:
#         header = "\n".join((key, "".join("-" for i in range(len(key))), ""))
#
#     if merge_within_sections and key not in doc_sections_that_cant_merge:
#         if child_sec is None:
#             body = prnt_sec
#         elif prnt_sec is None:
#             body = child_sec
#         else:
#             # only add same portion once
#             common = common_start(prnt_sec, child_sec)
#             n = len(common)
#             # remove additional 'new line' when append child
#             n = n + 1 if child_sec[n:].startswith('\n') else n
#             # ensure the common substring to be at least 10 character long
#             n = 0 if  n < 10 else n
#             body = '\n'.join((prnt_sec, child_sec[n:]))
#     else:
#         body = prnt_sec if child_sec is None else child_sec
#
#     return header + body
#
#
# def merge_all_sections(prnt_sctns, child_sctns, merge_within_sections=True):
#     """ Merge the doc-sections of the parent's and child's attribute into a single docstring.
#     Parameters
#     ----------
#     prnt_sctns: OrderedDict[str, Union[None,str]]
#     child_sctns: OrderedDict[str, Union[None,str]]
#     merge_within_sections: bool
#         merge parents and child docstring
#     Returns
#     -------
#     str
#         Output docstring of the merged docstrings.
#     """
#     doc = []
#
#     prnt_only_raises = prnt_sctns["Raises"] and not (
#         prnt_sctns["Returns"] or prnt_sctns["Yields"]
#     )
#     if prnt_only_raises and (child_sctns["Returns"] or child_sctns["Yields"]):
#         prnt_sctns["Raises"] = None
#
#     for key in prnt_sctns:
#         sect = merge_section(
#             key,
#             prnt_sctns[key],
#             child_sctns[key],
#             merge_within_sections=merge_within_sections
#         )
#         if sect is not None:
#             doc.append(sect)
#     return "\n\n".join(doc) if doc else None
#
#
# def merge_numpy_docs_dedup(prnt_doc=None, child_doc=None, merge_within_sections=True):
#     """ Merge two numpy-style docstrings into a single docstring.
#     Given the numpy-style docstrings from a parent and child's attributes, merge the docstring
#     sections such that the child's section is used, wherever present, otherwise the parent's
#     section is used.
#     Any whitespace that can be uniformly removed from a docstring's second line and onwards is
#     removed. Sections will be separated by a single blank line.
#     Parameters
#     ----------
#     prnt_doc: Optional[str]
#         The docstring from the parent.
#     child_doc: Optional[str]
#         The docstring from the child.
#     merge_within_sections: bool
#         merge parents and child docstring
#     Returns
#     -------
#     Union[str, None]
#         The merged docstring.
#     """
#     return merge_all_sections(
#         parse_numpy_doc(prnt_doc),
#         parse_numpy_doc(child_doc),
#         merge_within_sections=merge_within_sections
#     )



from collections import namedtuple
from enum import Enum


class PredictMethod(Enum):
    """
    The predict method for all of the stan template. Often used are mean and median.
    """

    MAP = "map"
    MEAN = "mean"
    MEDIAN = "median"
    FULL_SAMPLING = "full"


class PredictionKeys(Enum):
    """
    column names for the data frame of predicted result with decomposed components
    """

    TREND = "trend"
    SEASONALITY = "seasonality"
    REGRESSION = "regression"
    REGRESSOR = "regressor"
    PREDICTION = "prediction"


class EstimatorsKeys(Enum):
    """
    alias for all available estimator types when they are called under model wrapper functions
    """

    PyroSVI = "pyro-svi"
    StanMAP = "stan-map"
    StanMCMC = "stan-mcmc"
    CmdStanMAP = "cmdstan-map"
    CmdStanMCMC = "cmdstan-mcmc"


class TrainingMetaKeys(Enum):
    """
    training meta data dictionary processed under `Forecaster.fit()`
    """

    RESPONSE = "response"
    DATE_ARRAY = "date_array"
    NUM_OF_OBS = "num_of_obs"
    RESPONSE_SD = "response_sd"
    RESPONSE_MEAN = "response_mean"
    # START and END are in date-time format
    START = "training_start"
    END = "training_end"
    RESPONSE_COL = "response_col"
    DATE_COL = "date_col"


class PredictionMetaKeys(Enum):
    """
    prediction input meta data dictionary processed under `Forecaster.predict()`
    """

    DATE_ARRAY = "date_array"
    # FIXME: seems this is redudant? we can derive forecast range by input of data frame?
    # FIXME: or shall we cast this as zero when it is within training
    FUTURE_STEPS = "n_forecast_steps"
    # START and END are in date-time format
    START = "prediction_start"
    END = "prediction_end"
    START_INDEX = "start"
    END_INDEX = "end"
    PREDICTION_DF_LEN = "df_length"


class PlotLabels(Enum):
    """
    used in multiple prediction plots
    """

    # Also used in training_actual_response column name.
    TRAINING_ACTUAL_RESPONSE = "training_actual_response"
    PREDICTED_RESPONSE = "predicted_response"
    ACTUAL_RESPONSE = "actual_response"


class TimeSeriesSplitSchemeKeys(Enum):
    """hash table keys for the dictionary of back-test meta data"""

    MODEL = "model"
    TRAIN_START_DATE = "train_start_date"
    TRAIN_END_DATE = "train_end_date"
    TRAIN_IDX = "train_idx"
    TEST_IDX = "test_idx"
    # split scheme type
    SPLIT_TYPE_EXPANDING = "expanding"
    SPLIT_TYPE_ROLLING = "rolling"


class BacktestFitKeys(Enum):
    """column names of the dataframe used in the output from the backtest.BackTester.fit_predict() or any labels of
    the intermediate variables to generate such outcome dataframe
    """

    # labels for fitting process
    # note that the convention "_prediction" cannot be changed since it is also assumed in
    # all metric functions signature
    ACTUAL = "actual"
    PREDICTED = "prediction"
    DATE = "date"
    SPLIT_KEY = "split_key"
    TRAIN_FLAG = "training_data"
    TRAIN_ACTUAL = "train_actual"
    TRAIN_PREDICTED = "train_prediction"
    TEST_ACTUAL = "test_actual"
    TEST_PREDICTED = "test_prediction"
    # labels for scoring process
    METRIC_VALUES = "metric_values"
    METRIC_NAME = "metric_name"
    TRAIN_METRIC_FLAG = "is_training_metric"


class KTRTimePointPriorKeys(Enum):
    """hash table keys for the dictionary of back-test aggregation analysis result"""

    NAME = "name"
    PRIOR_START_TP_IDX = "prior_start_tp_idx"
    PRIOR_END_TP_IDX = "prior_end_tp_idx"
    PRIOR_MEAN = "prior_mean"
    PRIOR_SD = "prior_sd"
    PRIOR_REGRESSOR_COL = "prior_regressor_col"


# Defaults Values
DEFAULT_REGRESSOR_SIGN = "="
DEFAULT_REGRESSOR_BETA = 0
DEFAULT_REGRESSOR_SIGMA = 1.0

# beta coef columns
COEFFICIENT_DF_COLS = namedtuple(
    "coefficients_df_cols",
    ["REGRESSOR", "REGRESSOR_SIGN", "COEFFICIENT", "PROB_COEF_POS", "PROB_COEF_NEG"],
)("regressor", "regressor_sign", "coefficient", "Pr(coef >= 0)", "Pr(coef < 0)")


class CompiledStanModelPath:
    """
    the directory path for compliled stan models
    """

    PARENT = "orbit"
    CHILD = "stan"



from enum import Enum
import seaborn as sns
import matplotlib.colors as clr


class KTRPalette(Enum):
    """
    str
    """

    KNOTS_SEGMENT = "#276ef1"
    KNOTS_REGION = "#05A357"


class OrbitPalette(Enum):
    """
    str
    """

    # Qualitative Palette
    # single brand colors
    BLACK = "#000000"
    BLUE = "#276EF1"
    BLUE600 = "#174291"
    GREEN = "#05A357"
    GREEN600 = "#03582F"
    YELLOW = "#FFC043"
    YELLOW400 = "#FFC043"
    RED = "#E11900"
    BROWN = "#99644C"
    ORANGE = "#ED6E33"
    PURPLE = "#7356BF"
    WHITE = "#FFFFFF"


class OrbitColorMap(Enum):
    """
    matplotlib ColorMap
    """

    # Sequential Palette
    BLACK_GRADIENT = clr.LinearSegmentedColormap.from_list(
        "custom", [OrbitPalette.WHITE.value, OrbitPalette.BLACK.value], N=300
    )
    BLUE_GRADIENT = clr.LinearSegmentedColormap.from_list(
        "custom",
        [OrbitPalette.WHITE.value, OrbitPalette.BLUE.value, OrbitPalette.BLACK.value],
        N=300,
    )
    GREEN_GRADIENT = clr.LinearSegmentedColormap.from_list(
        "custom",
        [OrbitPalette.WHITE.value, OrbitPalette.GREEN.value, OrbitPalette.BLACK.value],
        N=300,
    )
    YELLOW_GRADIENT = clr.LinearSegmentedColormap.from_list(
        "custom",
        [OrbitPalette.WHITE.value, OrbitPalette.YELLOW.value, OrbitPalette.BLACK.value],
        N=300,
    )
    RED_GRADIENT = clr.LinearSegmentedColormap.from_list(
        "custom",
        [OrbitPalette.WHITE.value, OrbitPalette.RED.value, OrbitPalette.BLACK.value],
        N=300,
    )
    PURPLE_GRADIENT = clr.LinearSegmentedColormap.from_list(
        "custom",
        [OrbitPalette.WHITE.value, OrbitPalette.PURPLE.value, OrbitPalette.BLACK.value],
        N=300,
    )

    # Diverging Palette -  blue green yellow orange red
    RAINBOW = clr.LinearSegmentedColormap.from_list(
        "custom", ["#276EF1", "#05A357", "#FFC043", "#ED6E33", "#E11900"], N=300
    )


class PredictionPaletteClassic(Enum):
    """
    str
    """

    # black
    ACTUAL_OBS = "#000000"
    # blue
    PREDICTION_LINE = "#276EF1"
    PREDICTION_INTERVAL = "#276EF1"
    # black
    HOLDOUT_VERTICAL_LINE = "#000000"
    # yellow
    TEST_OBS = "#FFC043"



import pytest
import numpy as np
from copy import copy

from orbit.models import ETS
from orbit.constants.constants import PredictionKeys


@pytest.mark.parametrize("estimator", ["stan-map", "stan-mcmc"])
def test_base_ets_init(estimator):
    ets = ETS(estimator=estimator)

    is_fitted = ets.is_fitted()

    model_data_input = ets.get_training_data_input()
    model_param_names = ets._model.get_model_param_names()
    init_values = ets._model.get_init_values()

    # model is not yet fitted
    assert not is_fitted
    # should only be initialized and not set
    assert not model_data_input
    # model param names should already be set
    assert model_param_names
    # callable is not implemented yet
    assert not init_values


def test_ets_full_seasonal_fit(make_weekly_data):
    train_df, test_df, coef = make_weekly_data

    ets = ETS(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=52,
        num_warmup=50,
        num_sample=50,
        verbose=False,
        estimator="stan-mcmc",
    )
    ets.fit(train_df)

    init_values = ets._model.get_init_values()
    assert init_values["init_sea"].shape == (51,)

    predict_df = ets.predict(test_df, decompose=False)

    expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(ets._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ets._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("point_method", ["mean", "median"])
def test_ets_aggregated_seasonal_fit(make_weekly_data, point_method):
    train_df, test_df, coef = make_weekly_data

    ets = ETS(
        response_col="response",
        date_col="week",
        seasonality=52,
        num_warmup=50,
        num_sample=50,
        verbose=False,
        estimator="stan-mcmc",
        n_bootstrap_draws=1e4,
    )
    ets.fit(train_df, point_method=point_method)

    init_values = ets._model.get_init_values()
    assert init_values["init_sea"].shape == (51,)

    predict_df = ets.predict(test_df, decompose=False)

    expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(ets._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ets._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("n_bootstrap_draws", [None, -1, 1e4])
def test_ets_map_seasonal_fit(make_weekly_data, n_bootstrap_draws):
    train_df, test_df, coef = make_weekly_data

    ets = ETS(
        response_col="response",
        date_col="week",
        seasonality=52,
        verbose=False,
        estimator="stan-map",
        n_bootstrap_draws=n_bootstrap_draws,
    )

    ets.fit(train_df)
    init_values = ets._model.get_init_values()
    assert init_values["init_sea"].shape == (51,)

    predict_df = ets.predict(test_df)

    # if n_bootstrap_draws provided then produce prediction range
    if n_bootstrap_draws is not None and n_bootstrap_draws > 0:
        expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    else:
        expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(ets._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    p = ets.get_posterior_samples()
    assert len(p.keys()) == expected_num_parameters


def test_ets_non_seasonal_fit(make_weekly_data):
    train_df, test_df, coef = make_weekly_data

    ets = ETS(
        response_col="response",
        date_col="week",
        num_warmup=50,
        num_sample=50,
    )
    ets.fit(train_df)
    predict_df = ets.predict(test_df)

    expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(ets._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ets._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("prediction_percentiles", [None, [5, 10, 95]])
def test_full_prediction_percentiles(iclaims_training_data, prediction_percentiles):
    df = iclaims_training_data

    ets = ETS(
        response_col="claims",
        date_col="week",
        seasonality=52,
        num_warmup=50,
        num_sample=50,
        seed=8888,
        prediction_percentiles=prediction_percentiles,
    )

    if not prediction_percentiles:
        p_labels = ["_5", "", "_95"]
    else:
        p_labels = ["_5", "_10", "", "_95"]

    # test behaviors without decomposition
    ets.fit(df)
    predicted_df = ets.predict(df)
    expected_columns = ["week"] + ["prediction" + p for p in p_labels]
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]

    # test behaviors with decomposition
    predicted_df = ets.predict(df, decompose=True)
    plot_components = [
        "prediction",
        PredictionKeys.TREND.value,
        PredictionKeys.SEASONALITY.value,
    ]

    expected_columns = ["week"]
    for pc in plot_components:
        for p in p_labels:
            expected_columns.append(pc + p)
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]


@pytest.mark.parametrize("n_bootstrap_draws", [-1, 1e4])
@pytest.mark.parametrize("prediction_percentiles", [None, [5, 10, 95]])
def test_map_prediction_percentiles(
    iclaims_training_data, n_bootstrap_draws, prediction_percentiles
):
    df = iclaims_training_data

    ets = ETS(
        response_col="claims",
        date_col="week",
        seasonality=52,
        seed=8888,
        prediction_percentiles=prediction_percentiles,
        estimator="stan-map",
        n_bootstrap_draws=n_bootstrap_draws,
    )

    if n_bootstrap_draws > 0:
        if not prediction_percentiles:
            p_labels = ["_5", "", "_95"]
        else:
            p_labels = ["_5", "_10", "", "_95"]
    else:
        p_labels = [""]

    ets.fit(df)
    # test behaviors without decomposition
    predicted_df = ets.predict(df)

    expected_columns = ["week"] + ["prediction" + p for p in p_labels]

    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]

    # test behaviors with decomposition
    predicted_df = ets.predict(df, decompose=True)
    predicted_components = [
        "prediction",
        PredictionKeys.TREND.value,
        PredictionKeys.SEASONALITY.value,
    ]
    expected_columns = ["week"]

    for pc in predicted_components:
        for p in p_labels:
            expected_columns.append(pc + p)
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]


@pytest.mark.parametrize("estimator", ["stan-mcmc", "stan-map"])
def test_ets_missing(iclaims_training_data, estimator):
    df = iclaims_training_data
    missing_idx = np.array([10, 20, 30, 40, 41, 42, 43, 44, df.shape[0] - 1])
    df.loc[missing_idx, "claims"] = np.nan

    dlt = ETS(
        response_col="claims",
        date_col="week",
        seasonality=52,
        verbose=False,
        estimator=estimator,
    )

    dlt.fit(df)
    predicted_df = dlt.predict(df)
    if estimator == "stan-map":
        expected_columns = ["week", "prediction"]
    elif estimator == "stan-mcmc":
        expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]

    assert all(~np.isnan(predicted_df["prediction"]))
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]


@pytest.mark.parametrize("seasonality", [1, 52])
def test_ets_full_reproducibility(make_weekly_data, seasonality):
    train_df, test_df, coef = make_weekly_data

    ets1 = ETS(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        num_warmup=50,
        num_sample=50,
        estimator="stan-mcmc",
    )

    # first fit and predict
    ets1.fit(train_df)
    posteriors1 = ets1.get_posterior_samples()
    prediction1 = ets1.predict(test_df)

    # second fit and predict
    # note a new instance must be created to reset the seed
    # note both fit and predict contain random generation processes
    ets2 = ETS(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        num_warmup=50,
        num_sample=50,
        estimator="stan-mcmc",
    )

    ets2.fit(train_df)
    posteriors2 = ets2.get_posterior_samples()
    prediction2 = ets2.predict(test_df)

    # assert same posterior keys
    assert set(posteriors1.keys()) == set(posteriors2.keys())

    # assert posterior draws are reproducible
    for k, v in posteriors1.items():
        assert np.allclose(posteriors1[k], posteriors2[k])

    # assert prediction is reproducible
    assert prediction1.equals(prediction2)


@pytest.mark.parametrize("seasonality", [1, 52])
def test_ets_map_reproducibility(make_weekly_data, seasonality):
    train_df, test_df, coef = make_weekly_data

    ets1 = ETS(
        response_col="response",
        date_col="week",
        seasonality=seasonality,
        estimator="stan-map",
    )

    # first fit and predict
    ets1.fit(train_df)
    posteriors1 = ets1.get_point_posteriors()["map"]
    prediction1 = ets1.predict(test_df)

    # second fit and predict
    # note a new instance must be created to reset the seed
    # note both fit and predict contain random generation processes
    ets2 = ETS(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        estimator="stan-map",
    )

    ets2.fit(train_df)
    posteriors2 = ets2.get_point_posteriors()["map"]
    prediction2 = ets2.predict(test_df)

    # assert same posterior keys
    assert set(posteriors1.keys()) == set(posteriors2.keys())

    # assert posterior draws are reproducible
    for k, v in posteriors1.items():
        assert np.allclose(posteriors1[k], posteriors2[k])

    # assert prediction is reproducible
    assert np.allclose(
        prediction1["prediction"].values, prediction2["prediction"].values
    )


@pytest.mark.parametrize("estimator", ["stan-mcmc", "stan-map"])
@pytest.mark.parametrize("random_seed", [10, 100])
def test_ets_predict_seed(make_weekly_data, estimator, random_seed):
    train_df, test_df, coef = make_weekly_data
    args = {
        "response_col": "response",
        "date_col": "week",
        "seasonality": 52,
        "n_bootstrap_draws": 100,
        "verbose": False,
        "estimator": estimator,
    }

    if estimator == "stan-mcmc":
        args.update({"num_warmup": 50, "num_sample": 100})
    elif estimator == "pyro-svi":
        args.update({"num_steps": 10})

    ets = ETS(**args)
    ets.fit(train_df)
    predict_df1 = ets.predict(test_df, seed=random_seed)
    predict_df2 = ets.predict(test_df, seed=random_seed)

    assert all(predict_df1["prediction"].values == predict_df2["prediction"].values)


@pytest.mark.parametrize(
    "idx_range",
    [
        [0, 100],
        [0, 50],
        [50, 100],
        [50, 80],
        [50, 120],
        [100, 150],
    ],
    ids=[
        "train-start-to-train-end",
        "train-start-to-middle",
        "middle-to-train-end",
        "train-period-subset",
        "train-test-period-cross",
        "completely-test-period",
    ],
)
def test_ets_predict_range(make_weekly_data, idx_range):
    train_cut_off = 100
    base_df, _, _ = make_weekly_data
    train_df = base_df[:train_cut_off].reset_index(drop=True)
    predict_df = base_df[idx_range[0] : idx_range[1]].reset_index(drop=True)

    ets = ETS(
        response_col="response",
        date_col="week",
        seasonality=52,
        verbose=False,
        estimator="stan-map",
    )
    ets.fit(train_df)
    ets.predict(predict_df)



from copy import copy
import pytest
import numpy as np

from orbit.models import DLT
from orbit.constants.constants import PredictionKeys
from orbit.exceptions import ModelException, PredictionException


@pytest.mark.parametrize("estimator", ["stan-map", "stan-mcmc"])
def test_base_dlt_init(estimator):
    dlt = DLT(estimator=estimator)

    is_fitted = dlt.is_fitted()

    model_data_input = dlt.get_training_data_input()
    model_param_names = dlt._model.get_model_param_names()
    init_values = dlt._model.get_init_values()

    # model is not yet fitted
    assert not is_fitted
    # should only be initialized and not set
    assert not model_data_input
    # model param names should already be set
    assert model_param_names
    # callable is not implemented yet
    assert not init_values


@pytest.mark.parametrize("estimator", ["stan-mcmc"])
def test_dlt_full_univariate(make_weekly_data, estimator):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=52,
        num_warmup=50,
        verbose=False,
        estimator=estimator,
    )

    dlt.fit(train_df)

    init_values = dlt._model.get_init_values()
    assert init_values["init_sea"].shape == (51,)

    predict_df = dlt.predict(test_df)

    expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(dlt._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(dlt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("estimator", ["stan-mcmc"])
@pytest.mark.parametrize("point_method", ["mean", "median"])
def test_dlt_aggregated_univariate(make_weekly_data, estimator, point_method):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        seasonality=52,
        num_warmup=50,
        verbose=False,
        estimator=estimator,
    )

    dlt.fit(train_df, point_method=point_method)

    init_values = dlt._model.get_init_values()
    assert init_values["init_sea"].shape == (51,)

    predict_df = dlt.predict(test_df)

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(dlt._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(dlt._posterior_samples) == expected_num_parameters


def test_dlt_map_univariate(make_weekly_data):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        seasonality=52,
        num_warmup=50,
        verbose=False,
        estimator="stan-map",
    )

    dlt.fit(train_df)

    init_values = dlt._model.get_init_values()
    assert init_values["init_sea"].shape == (51,)

    predict_df = dlt.predict(test_df)

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(dlt._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(dlt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("estimator", ["stan-mcmc"])
def test_dlt_non_seasonal_fit(make_weekly_data, estimator):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        estimator=estimator,
        num_warmup=50,
    )

    dlt.fit(train_df)
    predict_df = dlt.predict(test_df)

    expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (51, len(expected_columns))
    expected_num_parameters = len(dlt._model.get_model_param_names()) + 1

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(dlt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+", "+", "+", "+"],
        ["-", "-", "-", "-", "-", "-"],
        ["=", "=", "=", "=", "=", "="],
    ],
    ids=["positive_only", "negative_only", "regular_only"],
)
@pytest.mark.parametrize(
    "invalid_input",
    [np.nan, np.inf, -1 * np.inf],
    ids=["nan", "infinite", "neg-infinite"],
)
def test_invalid_regressor(make_weekly_data, regressor_signs, invalid_input):
    train_df, test_df, coef = make_weekly_data
    regressor_col = train_df.columns.tolist()[2:]
    # make invalid values
    train_df[regressor_col[0]][36] = invalid_input
    expected_flag = False
    try:
        dlt = DLT(
            response_col="response",
            date_col="week",
            regressor_col=regressor_col,
            regressor_sign=regressor_signs,
            prediction_percentiles=[5, 95],
            seasonality=52,
            num_warmup=50,
            verbose=False,
            estimator="stan-map",
        )
        dlt.fit(train_df)
    except ModelException:
        expected_flag = True

    assert expected_flag


@pytest.mark.parametrize(
    "invalid_input",
    [np.nan, np.inf, -1 * np.inf],
    ids=["nan", "infinite", "neg-infinite"],
)
def test_invalid_predict_regressor(make_weekly_data, invalid_input):
    train_df, test_df, coef = make_weekly_data
    regressor_col = train_df.columns.tolist()[2:]
    # make invalid values
    test_df[regressor_col[0]][3] = invalid_input
    expected_flag = False
    try:
        dlt = DLT(
            response_col="response",
            date_col="week",
            regressor_col=regressor_col,
            prediction_percentiles=[5, 95],
            seasonality=52,
            num_warmup=50,
            verbose=False,
            estimator="stan-map",
        )
        dlt.fit(train_df)
        dlt.predict(test_df)
    except PredictionException:
        expected_flag = True

    assert expected_flag


@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+", "+", "+", "+"],
        ["-", "-", "-", "-", "-", "-"],
        ["=", "=", "=", "=", "=", "="],
        ["+", "=", "+", "=", "-", "-"],
    ],
    ids=["positive_only", "negative_only", "regular_only", "mixed_signs"],
)
def test_dlt_full_with_regression(make_weekly_data, regressor_signs):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        regressor_sign=regressor_signs,
        prediction_percentiles=[5, 95],
        seasonality=52,
        num_warmup=50,
        verbose=False,
        estimator="stan-mcmc",
    )

    dlt.fit(train_df)
    init_values = dlt._model.get_init_values()
    assert init_values["init_sea"].shape == (51,)

    if regressor_signs.count("+") > 0:
        assert init_values["pr_beta"].shape == (regressor_signs.count("+"),)
    if regressor_signs.count("-") > 0:
        assert init_values["nr_beta"].shape == (regressor_signs.count("-"),)
    if regressor_signs.count("=") > 0:
        assert init_values["rr_beta"].shape == (regressor_signs.count("="),)

    predict_df = dlt.predict(test_df)

    regression_out = dlt.get_regression_coefs()
    num_regressors = regression_out.shape[0]

    expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (51, len(expected_columns))
    expected_regression_shape = (6, 7)

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert regression_out.shape == expected_regression_shape
    assert num_regressors == len(train_df.columns.tolist()[2:])

    assert np.sum(regression_out["coefficient"].values >= 0) <= regressor_signs.count(
        "+"
    ) + regressor_signs.count("=")
    assert np.sum(regression_out["coefficient"].values <= 0) <= regressor_signs.count(
        "-"
    ) + regressor_signs.count("=")


@pytest.mark.parametrize("estimator", ["stan-mcmc"])
@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+", "+", "+", "+"],
        ["-", "-", "-", "-", "-", "-"],
        ["=", "=", "=", "=", "=", "="],
        ["+", "=", "+", "=", "-", "-"],
    ],
    ids=["positive_only", "negative_only", "regular_only", "mixed_signs"],
)
@pytest.mark.parametrize("point_method", ["mean", "median"])
def test_dlt_aggregated_with_regression(
    make_weekly_data, estimator, regressor_signs, point_method
):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        regressor_sign=regressor_signs,
        seasonality=52,
        num_warmup=50,
        verbose=False,
        estimator=estimator,
    )

    dlt.fit(train_df, point_method=point_method)
    predict_df = dlt.predict(test_df)

    regression_out = dlt.get_regression_coefs()
    num_regressors = regression_out.shape[0]

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_regression_shape = (6, 7)

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert regression_out.shape == expected_regression_shape
    assert num_regressors == len(train_df.columns.tolist()[2:])

    predict_df = dlt.predict(test_df, decompose=True)
    assert any(predict_df["regression"].values)


@pytest.mark.parametrize(
    "global_trend_option", ["linear", "loglinear", "logistic", "flat"]
)
def test_dlt_map_global_trend(make_weekly_data, global_trend_option):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        seasonality=52,
        global_trend_option=global_trend_option,
        estimator="stan-map",
    )

    dlt.fit(train_df)
    predict_df = dlt.predict(test_df)

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns


@pytest.mark.parametrize(
    # true signs
    # coef=[0.2, 0.1, 0.3, 0.15, -0.2, -0.1]
    "regressor_signs",
    [
        ["+", "+", "+", "+", "=", "="],
        ["=", "=", "=", "=", "-", "-"],
        ["+", "+", "+", "+", "-", "-"],
        ["+", "+", "+", "=", "=", "-"],
    ],
    ids=["positive_mixed", "negative_mixed", "positive_negative", "mixed"],
)
def test_dlt_mixed_signs_and_order(make_weekly_data, regressor_signs):
    df, _, _ = make_weekly_data
    raw_regressor_col = ["a", "b", "c", "d", "e", "f"]
    new_regressor_col = [raw_regressor_col[idx] for idx in [1, 2, 0, 5, 3, 4]]
    new_regressor_signs = [regressor_signs[idx] for idx in [1, 2, 0, 5, 3, 4]]

    # mixing ordering of cols in df of prediction
    new_df = df[["response", "week"] + new_regressor_col]

    dlt = DLT(
        response_col="response",
        date_col="week",
        regressor_col=raw_regressor_col,
        regressor_sign=regressor_signs,
        seasonality=52,
        seed=8888,
        num_warmup=4000,
        num_sample=4000,
        estimator="stan-mcmc",
    )
    dlt.fit(df)
    coef_df = dlt.get_regression_coefs().set_index("regressor")
    coefs = coef_df.loc[raw_regressor_col, "coefficient"].values

    predicted_df_v1 = dlt.predict(df, decompose=True)
    predicted_df_v2 = dlt.predict(new_df, decompose=True)

    # mixing ordering of signs
    dlt_new = DLT(
        response_col="response",
        date_col="week",
        regressor_col=new_regressor_col,
        regressor_sign=new_regressor_signs,
        seasonality=52,
        seed=8888,
        num_warmup=4000,
        num_sample=4000,
        estimator="stan-mcmc",
    )
    dlt_new.fit(df)
    new_coef_df = dlt_new.get_regression_coefs().set_index("regressor")
    new_coefs = new_coef_df.loc[raw_regressor_col, "coefficient"].values

    # coefficients should be as close as  <= 0.01
    assert np.allclose(coefs, new_coefs, atol=1e-2)

    predicted_df_v3 = dlt_new.predict(df, decompose=True)
    predicted_df_v4 = dlt_new.predict(new_df, decompose=True)

    # relative ratio of regression comp to acutal
    pred_v1 = predicted_df_v1["regression"].values / df["response"].values
    pred_v2 = predicted_df_v2["regression"].values / df["response"].values
    pred_v3 = predicted_df_v3["regression"].values / df["response"].values
    pred_v4 = predicted_df_v4["regression"].values / df["response"].values

    # they should be all identical; ordering of signs or columns in prediction df should show not material difference
    # exclude the first one which is used for initialization and hence more unstable
    # exchange columns in prediction should have minimal variance
    assert np.allclose(pred_v1[1:], pred_v2[1:], atol=1e-3)
    assert np.allclose(pred_v3[1:], pred_v4[1:], atol=1e-3)
    # exchange columns in input of fitting should have less variance
    assert np.allclose(pred_v1[1:], pred_v3[1:], atol=1e-1)


@pytest.mark.parametrize("prediction_percentiles", [None, [5, 10, 95]])
def test_dlt_prediction_percentiles(iclaims_training_data, prediction_percentiles):
    df = iclaims_training_data

    dlt = DLT(
        response_col="claims",
        date_col="week",
        seasonality=52,
        num_warmup=50,
        num_sample=50,
        seed=8888,
        prediction_percentiles=prediction_percentiles,
        estimator="stan-mcmc",
    )

    if not prediction_percentiles:
        p_labels = ["_5", "", "_95"]
    else:
        p_labels = ["_5", "_10", "", "_95"]

    dlt.fit(df)
    predicted_df = dlt.predict(df)
    expected_columns = ["week"] + ["prediction" + p for p in p_labels]
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]

    predicted_df = dlt.predict(df, decompose=True)
    predicted_components = [
        "prediction",
        PredictionKeys.TREND.value,
        PredictionKeys.SEASONALITY.value,
        PredictionKeys.REGRESSION.value,
    ]

    expected_columns = ["week"]
    for pc in predicted_components:
        for p in p_labels:
            expected_columns.append(pc + p)
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]


@pytest.mark.parametrize("estimator", ["stan-mcmc"])
@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+", "+", "+", "+"],
        ["=", "=", "=", "=", "=", "="],
        ["-", "-", "-", "-", "-", "-"],
        ["+", "=", "+", "=", "+", "+"],
        ["-", "=", "-", "=", "-", "="],
        ["+", "=", "+", "=", "-", "-"],
    ],
    ids=[
        "positive_only",
        "regular_only",
        "negative_only",
        "positive_mixed",
        "negative_mixed",
        "mixed_signs",
    ],
)
@pytest.mark.parametrize("seasonality", [1, 52])
def test_dlt_full_reproducibility(
    make_weekly_data, estimator, regressor_signs, seasonality
):
    train_df, test_df, coef = make_weekly_data

    dlt_first = DLT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        regressor_sign=regressor_signs,
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        num_warmup=50,
        verbose=False,
        estimator=estimator,
    )

    # first fit and predict
    dlt_first.fit(train_df)
    posteriors_first = copy(dlt_first._posterior_samples)
    predict_df_first = dlt_first.predict(test_df)
    regression_out_first = dlt_first.get_regression_coefs()

    # second fit and predict
    # note a new instance must be created to reset the seed
    # note both fit and predict contain random generation processes
    dlt_second = DLT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        regressor_sign=regressor_signs,
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        num_warmup=50,
        verbose=False,
        estimator=estimator,
    )

    dlt_second.fit(train_df)
    posteriors_second = copy(dlt_second._posterior_samples)
    predict_df_second = dlt_second.predict(test_df)
    regression_out_second = dlt_second.get_regression_coefs()

    # assert same posterior keys
    assert set(posteriors_first.keys()) == set(posteriors_second.keys())

    # assert posterior draws are reproducible
    for k, v in posteriors_first.items():
        assert np.allclose(posteriors_first[k], posteriors_second[k])

    # assert identical regression columns
    # this is also checked in posterior samples, but an extra layer just in case
    # since this one very commonly retrieved by end users
    assert regression_out_first.equals(regression_out_second)

    # assert prediction is reproducible
    assert predict_df_first.equals(predict_df_second)


@pytest.mark.parametrize("seasonality", [1, 52])
def test_dlt_map_reproducibility(make_weekly_data, seasonality):
    train_df, test_df, coef = make_weekly_data

    dlt1 = DLT(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        estimator="stan-map",
    )

    # first fit and predict
    dlt1.fit(train_df)
    posteriors1 = copy(dlt1._point_posteriors["map"])
    prediction1 = dlt1.predict(test_df)

    # second fit and predict
    # note a new instance must be created to reset the seed
    # note both fit and predict contain random generation processes
    dlt2 = DLT(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        estimator="stan-map",
    )

    dlt2.fit(train_df)
    posteriors2 = copy(dlt2._point_posteriors["map"])
    prediction2 = dlt2.predict(test_df)

    # assert same posterior keys
    assert set(posteriors1.keys()) == set(posteriors2.keys())

    # assert posterior draws are reproducible
    for k, v in posteriors1.items():
        assert np.allclose(posteriors1[k], posteriors2[k])

    # assert prediction is reproducible
    assert np.allclose(
        prediction1["prediction"].values, prediction2["prediction"].values
    )


@pytest.mark.parametrize("regression_penalty", ["fixed_ridge", "lasso", "auto_ridge"])
def test_dlt_regression_penalty(make_weekly_data, regression_penalty):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        regression_penalty=regression_penalty,
        seasonality=52,
        num_warmup=50,
        verbose=False,
        estimator="stan-map",
    )

    dlt.fit(train_df)
    predict_df = dlt.predict(test_df)

    regression_out = dlt.get_regression_coefs()
    num_regressors = regression_out.shape[0]

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_regression_shape = (6, 3)

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert regression_out.shape == expected_regression_shape
    assert num_regressors == len(train_df.columns.tolist()[2:])


@pytest.mark.parametrize("level_sm_input", [0.0001, 0.5, 1.0])
@pytest.mark.parametrize("seasonality_sm_input", [0.0, 0.5, 1.0])
@pytest.mark.parametrize("slope_sm_input", [0.0, 0.5, 1.0])
def test_dlt_fixed_sm_input(
    make_weekly_data, level_sm_input, seasonality_sm_input, slope_sm_input
):
    train_df, test_df, coef = make_weekly_data

    dlt = DLT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        level_sm_input=level_sm_input,
        seasonality_sm_input=seasonality_sm_input,
        slope_sm_input=slope_sm_input,
        seasonality=52,
        num_warmup=50,
        verbose=False,
        estimator="stan-map",
    )

    dlt.fit(train_df)
    predict_df = dlt.predict(test_df)

    regression_out = dlt.get_regression_coefs()
    num_regressors = regression_out.shape[0]

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_regression_shape = (6, 3)

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert regression_out.shape == expected_regression_shape
    assert num_regressors == len(train_df.columns.tolist()[2:])


@pytest.mark.parametrize("estimator", ["stan-mcmc", "stan-map"])
def test_dlt_missing(iclaims_training_data, estimator):
    df = iclaims_training_data
    missing_idx = np.array([10, 20, 30, 40, 41, 42, 43, 44, df.shape[0] - 1])
    df.loc[missing_idx, "claims"] = np.nan

    dlt = DLT(
        response_col="claims",
        date_col="week",
        seasonality=52,
        verbose=False,
        estimator=estimator,
    )

    dlt.fit(df)
    predicted_df = dlt.predict(df)
    if estimator == "stan-map":
        expected_columns = ["week", "prediction"]
    elif estimator == "stan-mcmc":
        expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]

    assert all(~np.isnan(predicted_df["prediction"]))
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]


def test_dlt_map_single_regressor(iclaims_training_data):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])
    regressor_col = ["trend.unemploy"]

    dlt = DLT(
        response_col="claims",
        date_col="week",
        regressor_col=regressor_col,
        seasonality=52,
        seed=8888,
        estimator="stan-map",
    )
    dlt.fit(df)
    predicted_df = dlt.predict(df)

    expected_num_parameters = len(dlt._model.get_model_param_names()) + 1
    expected_columns = ["week", "prediction"]

    assert predicted_df.shape[0] == df.shape[0]
    assert predicted_df.columns.tolist() == expected_columns
    assert len(dlt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("estimator", ["stan-mcmc"])
@pytest.mark.parametrize("keep_samples", [True, False])
@pytest.mark.parametrize("point_method", ["mean", "median"])
def test_dlt_is_fitted(iclaims_training_data, estimator, keep_samples, point_method):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])
    regressor_col = ["trend.unemploy"]

    dlt = DLT(
        response_col="claims",
        date_col="week",
        regressor_col=regressor_col,
        seasonality=52,
        seed=8888,
        num_warmup=50,
        num_sample=50,
        verbose=False,
        estimator=estimator,
    )

    dlt.fit(df, keep_samples=keep_samples, point_method=point_method)
    is_fitted = dlt.is_fitted()

    # still True when keep_samples is False
    assert is_fitted


@pytest.mark.parametrize("estimator", ["stan-mcmc", "stan-map"])
@pytest.mark.parametrize("random_seed", [10, 100])
def test_dlt_predict_seed(make_weekly_data, estimator, random_seed):
    train_df, test_df, coef = make_weekly_data
    args = {
        "response_col": "response",
        "date_col": "week",
        "seasonality": 52,
        "n_bootstrap_draws": 100,
        "verbose": False,
        "estimator": estimator,
    }

    if estimator == "stan-mcmc":
        args.update({"num_warmup": 50, "num_sample": 100})
    elif estimator == "pyro-svi":
        args.update({"num_steps": 10})

    dlt = DLT(**args)
    dlt.fit(train_df)
    predict_df1 = dlt.predict(test_df, seed=random_seed)
    predict_df2 = dlt.predict(test_df, seed=random_seed)

    assert all(predict_df1["prediction"].values == predict_df2["prediction"].values)


@pytest.mark.parametrize(
    "idx_range",
    [
        [0, 100],
        [0, 50],
        [50, 100],
        [50, 80],
        [50, 120],
        [100, 150],
    ],
    ids=[
        "train-start-to-train-end",
        "train-start-to-middle",
        "middle-to-train-end",
        "train-period-subset",
        "train-test-period-cross",
        "completely-test-period",
    ],
)
def test_dlt_predict_range(make_weekly_data, idx_range):
    train_cut_off = 100
    base_df, _, _ = make_weekly_data
    train_df = base_df[:train_cut_off].reset_index(drop=True)
    predict_df = base_df[idx_range[0] : idx_range[1]].reset_index(drop=True)

    dlt = DLT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        seasonality=52,
        verbose=False,
        estimator="stan-map",
    )
    dlt.fit(train_df)
    dlt.predict(predict_df)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath(".."))
import matplotlib
import orbit

# -- Project information -----------------------------------------------------
master_doc = "index"

project = "orbit"
copyright = "2020, Uber Technologies, Inc"
author = "Edwin Ng, Steve Yang, Zhishi Wang, Yifeng Wu, Jing Pan"

# The short X.Y version.
version = orbit.__version__
# The full version, including alpha/beta/rc tags
release = orbit.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    # 'sphinx.ext.autosummary',
    # 'sphinx.ext.doctest',
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
    # 'sphinx.ext.viewcode',
    "sphinx.ext.napoleon",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "nbsphinx",
]

# autodoc_mock_imports = [
#     'pandas', 'torch', 'cmdstanpy', 'tqdm', 'matplotlib.pyplot'
#     'pyro', 'matplotlib', 'seaborn', 'scipy', 'sklearn']

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
# source_suffix = ['.rst', '.ipynb']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

html_context = {
    "display_github": True,
    "github_user": "uber",
    "github_repo": "orbit",
    "github_version": "master/docs/",
}

# logo
html_logo = "img/orbit-logo-black.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    # 'logo_only': False,
    "navigation_depth": 1,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_style = "css/orbit.css"


html_favicon = "img/favicon/favicon.ico"

napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True

nbsphinx_kernel_name = "python3"
nbsphinx_timeout = 60



import pytest
import numpy as np
import pandas as pd

from orbit.models import KTR
from orbit.diagnostics.metrics import smape

SMAPE_TOLERANCE = 0.2


@pytest.mark.parametrize("make_daily_data", [({"seasonality": None})], indirect=True)
def test_ktr_basic(make_daily_data):
    train_df, _, _ = make_daily_data

    ktr = KTR(
        response_col="response",
        date_col="date",
        estimator="pyro-svi",
        num_steps=100,
        num_sample=100,
        n_bootstrap_draws=-1,
    )

    ktr.fit(train_df)
    predict_df = ktr.predict(train_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (train_df.shape[0], len(expected_columns))
    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert np.all(np.isfinite(predict_df["prediction"].values))


@pytest.mark.parametrize(
    "seas_segments",
    [
        pytest.param(0, id="0-seas_segement"),
        pytest.param(1, id="2-seas_segement"),
        pytest.param(5, id="5-seas_segement"),
    ],
)
@pytest.mark.parametrize(
    "make_daily_data, seasonality",
    [
        pytest.param(({"seasonality": "single"}), [365.25], id="single_seasonality"),
        pytest.param(({"seasonality": "dual"}), [7, 365.25], id="dual_seasonality"),
    ],
    indirect=["make_daily_data"],
)
def test_ktr_seasonality(make_daily_data, seasonality, seas_segments):
    train_df, _, _ = make_daily_data

    ktr = KTR(
        response_col="response",
        date_col="date",
        estimator="pyro-svi",
        seasonality=seasonality,
        seasonality_segments=seas_segments,
        num_steps=100,
        num_sample=100,
        n_bootstrap_draws=-1,
    )

    ktr.fit(train_df)
    predict_df = ktr.predict(train_df, decompose=True)
    expected_columns = [
        "date",
        "prediction_5",
        "prediction",
        "prediction_95",
        "trend_5",
        "trend",
        "trend_95",
        "regression_5",
        "regression",
        "regression_95",
    ]

    if len(seasonality) == 1:
        expected_columns += [
            "seasonality_365.25_5",
            "seasonality_365.25",
            "seasonality_365.25_95",
        ]
    if len(seasonality) == 2:
        expected_columns += [
            "seasonality_7_5",
            "seasonality_7",
            "seasonality_7_95",
            "seasonality_365.25_5",
            "seasonality_365.25",
            "seasonality_365.25_95",
        ]
    expected_shape = (train_df.shape[0], len(expected_columns))
    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert np.all(np.isfinite(predict_df["prediction"].values))
    smape_val = smape(train_df["response"].values, predict_df["prediction"].values)
    assert smape_val <= SMAPE_TOLERANCE


@pytest.mark.parametrize("regressor_col", [None, ["a", "b", "c"]])
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "dual", "with_coef": True})], indirect=True
)
def test_ktr_regression(make_daily_data, regressor_col):
    train_df, test_df, coef = make_daily_data

    ktr = KTR(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        regressor_col=regressor_col,
        estimator="pyro-svi",
        num_steps=100,
        num_sample=100,
        n_bootstrap_draws=-1,
    )

    ktr.fit(train_df)
    predict_df = ktr.predict(test_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (364, len(expected_columns))

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert np.all(np.isfinite(predict_df["prediction"].values))


@pytest.mark.parametrize(
    "regression_knot_dates",
    [pd.date_range(start="2016-03-01", end="2019-01-01", freq="3M")],
)
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "dual", "with_coef": True})], indirect=True
)
def test_ktrx_coef_knot_dates(make_daily_data, regression_knot_dates):
    train_df, test_df, coef = make_daily_data

    ktr = KTR(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        regressor_col=["a", "b", "c"],
        regression_knot_dates=regression_knot_dates,
        estimator="pyro-svi",
        num_steps=100,
        num_sample=100,
        n_bootstrap_draws=-1,
    )

    ktr.fit(train_df)
    predict_df = ktr.predict(test_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktr.get_posterior_samples()) == expected_num_parameters
    assert np.all(np.isfinite(predict_df["prediction"].values))


@pytest.mark.parametrize("regression_knot_distance", [90, 120])
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "dual", "with_coef": True})], indirect=True
)
def test_ktrx_coef_knot_distance(make_daily_data, regression_knot_distance):
    train_df, test_df, coef = make_daily_data

    ktr = KTR(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        regressor_col=["a", "b", "c"],
        regression_knot_distance=regression_knot_distance,
        estimator="pyro-svi",
        num_steps=100,
        num_sample=100,
        n_bootstrap_draws=-1,
    )

    ktr.fit(train_df)
    predict_df = ktr.predict(test_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktr._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize(
    "regressor_signs",
    [["+", "+", "+"], ["-", "-", "-"], ["=", "=", "="], ["+", "=", "-"]],
    ids=["positive_only", "negative_only", "regular_only", "mixed_signs"],
)
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "dual", "with_coef": True})], indirect=True
)
def test_ktrx_regressor_sign(make_daily_data, regressor_signs):
    train_df, test_df, coef = make_daily_data

    ktr = KTR(
        response_col="response",
        date_col="date",
        regressor_col=["a", "b", "c"],
        regressor_sign=regressor_signs,
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        estimator="pyro-svi",
        num_steps=100,
        num_sample=100,
        n_bootstrap_draws=-1,
    )

    ktr.fit(train_df)
    predict_df = ktr.predict(test_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktr._posterior_samples) == expected_num_parameters

    coef_df = ktr.get_regression_coefs()
    for i, sign in enumerate(regressor_signs):
        if sign == "+":
            assert all(np.sign(coef_df.iloc[:, 1 + i]) == 1)
        elif sign == "-":
            assert all(np.sign(coef_df.iloc[:, 1 + i]) == -1)


@pytest.mark.parametrize(
    "coef_prior_list",
    [
        [
            {
                "name": "test1",
                "prior_start_tp_idx": 100,
                "prior_end_tp_idx": 120,
                "prior_mean": [0.25, 0.35],
                "prior_sd": [0.1, 0.2],
                "prior_regressor_col": ["a", "b"],
            }
        ]
    ],
)
@pytest.mark.parametrize(
    "make_daily_data", [({"seasonality": "dual", "with_coef": True})], indirect=True
)
def test_ktrx_prior_ingestion(make_daily_data, coef_prior_list):
    train_df, test_df, coef = make_daily_data

    ktr = KTR(
        response_col="response",
        date_col="date",
        seasonality=[7, 365.25],
        seasonality_fs_order=[2, 5],
        regressor_col=["a", "b", "c"],
        coef_prior_list=coef_prior_list,
        estimator="pyro-svi",
        num_steps=100,
        num_sample=100,
        n_bootstrap_draws=-1,
    )

    ktr.fit(train_df)
    predict_df = ktr.predict(test_df)

    expected_columns = ["date", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (364, len(expected_columns))
    expected_num_parameters = 7

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(ktr._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize(
    "regression_segments",
    [
        pytest.param(0, id="0-regression_segments"),
        pytest.param(1, id="1-regression_segments"),
        pytest.param(5, id="5-regression_segments"),
    ],
)
@pytest.mark.parametrize(
    "regressor_col",
    [
        pytest.param(["trend.unemploy"], id="single-regressor"),
        pytest.param(
            ["trend.unemploy", "trend.filling", "trend.job"], id="multiple-regressor"
        ),
    ],
)
def test_ktr_regressor_sizes(iclaims_training_data, regression_segments, regressor_col):
    # this iclaims dataset is raw; the one under load_data is log-transformed
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])

    ktr = KTR(
        date_col="week",
        response_col="claims",
        regressor_col=regressor_col,
        seasonality=[52],
        seasonality_fs_order=[3],
        level_knot_scale=0.1,
        level_segments=10,
        regression_segments=regression_segments,
        regression_rho=0.15,
        # pyro optimization parameters
        seed=8888,
        num_steps=100,
        num_sample=100,
    )
    ktr.fit(df)
    ktr.predict(df)
    coef_df = ktr.get_regression_coefs()
    knot_df = ktr.get_regression_coef_knots()

    expected_columns_coef = ["week"] + regressor_col
    expected_columns_knot = ["week", "step"] + regressor_col

    assert coef_df.columns.tolist() == expected_columns_coef
    assert knot_df.columns.tolist() == expected_columns_knot
    assert coef_df.shape[0] == df.shape[0]
    assert knot_df.shape[0] == regression_segments + 1



import pkg_resources
import functools
import logging
from matplotlib import pyplot as plt

STYLE_FILE_NAME = "plot_style"


def get_orbit_style():
    path = pkg_resources.resource_filename(
        "orbit", "stylelib/{}.mplstyle".format(STYLE_FILE_NAME)
    )
    return path


def orbit_style_decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # check if arg is specified; if not, set True as default
        if "use_orbit_style" in kwargs.keys():
            use_orbit_style = kwargs["use_orbit_style"]
            del kwargs["use_orbit_style"]
        else:
            use_orbit_style = True

        # use orbit style plot if it is set to be True
        if use_orbit_style:
            orbit_style_path = get_orbit_style()
            try:
                with plt.style.context(orbit_style_path):
                    return func(*args, **kwargs)
            except:
                logging.info(
                    "Cannot find path:{}. Use default plot style.".format(
                        orbit_style_path
                    )
                )
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return wrapper



from copy import copy
import pytest
import numpy as np

from orbit.models import LGT
from orbit.constants.constants import PredictionKeys


@pytest.mark.parametrize("estimator", ["stan-map", "stan-mcmc"])
def test_base_lgt_init(estimator):
    lgt = LGT(estimator=estimator)

    is_fitted = lgt.is_fitted()

    model_data_input = lgt.get_training_data_input()
    model_param_names = lgt._model.get_model_param_names()
    init_values = lgt._model.get_init_values()

    # model is not yet fitted
    assert not is_fitted
    # should only be initialized and not set
    assert not model_data_input
    # model param names should already be set
    assert model_param_names
    # callable is not implemented yet
    assert not init_values


@pytest.mark.parametrize("seasonality", [None, 52])
@pytest.mark.parametrize("store_prediction_array", [True, False])
@pytest.mark.parametrize("n_bootstrap_draws", [None, 50])
@pytest.mark.parametrize("point_method", [None, "median"])
@pytest.mark.parametrize("estimator", ["stan-mcmc", "pyro-svi"])
def test_lgt_full_fit(
    make_weekly_data,
    seasonality,
    estimator,
    n_bootstrap_draws,
    store_prediction_array,
    point_method,
):
    train_df, test_df, coef = make_weekly_data
    args = {
        "response_col": "response",
        "date_col": "week",
        "prediction_percentiles": [5, 95],
        "seasonality": seasonality,
        "n_bootstrap_draws": n_bootstrap_draws,
        "verbose": False,
        "estimator": estimator,
    }

    if estimator == "stan-mcmc":
        args.update({"num_warmup": 50, "num_sample": 100})
    elif estimator == "pyro-svi":
        args.update({"num_steps": 10})

    # if seasonality == 52:
    #     expected_num_parameters += 2

    lgt = LGT(**args)
    lgt.fit(train_df, point_method=point_method)
    if estimator == "stan-mcmc":
        expected_num_parameters = len(lgt._model.get_model_param_names()) + 1
    elif estimator == "pyro-svi":
        expected_num_parameters = len(lgt._model.get_model_param_names())

    init_values = lgt._model.get_init_values()
    if seasonality:
        assert init_values["init_sea"].shape == (51,)
    else:
        assert init_values is None

    predict_df = lgt.predict(test_df, store_prediction_array=store_prediction_array)

    _ = lgt.get_prediction_meta()
    _ = lgt.get_training_metrics()

    if not n_bootstrap_draws and point_method:
        expected_columns = ["week", "prediction"]
        expected_shape = (51, len(expected_columns))
    else:
        expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
        expected_shape = (51, len(expected_columns))

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(lgt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("seasonality", [None, 52])
@pytest.mark.parametrize("estimator", ["stan-mcmc", "pyro-svi"])
@pytest.mark.parametrize("point_method", ["mean", "median"])
def test_lgt_aggregated_fit(make_weekly_data, seasonality, estimator, point_method):
    train_df, test_df, coef = make_weekly_data
    args = {
        "response_col": "response",
        "date_col": "week",
        "prediction_percentiles": [5, 95],
        "seasonality": seasonality,
        "verbose": False,
        "estimator": estimator,
    }
    if estimator == "stan-mcmc":
        args.update({"num_warmup": 50, "num_sample": 50})
    elif estimator == "pyro-svi":
        args.update({"num_steps": 10, "num_sample": 50})

    lgt = LGT(**args)
    lgt.fit(train_df, point_method=point_method)
    if estimator == "stan-mcmc":
        expected_num_parameters = len(lgt._model.get_model_param_names()) + 1
    elif estimator == "pyro-svi":
        expected_num_parameters = len(lgt._model.get_model_param_names())

    init_values = lgt._model.get_init_values()
    if seasonality:
        assert init_values["init_sea"].shape == (51,)
    else:
        assert init_values is None

    predict_df = lgt.predict(test_df)
    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(lgt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("seasonality", [None, 52])
@pytest.mark.parametrize("estimator", ["stan-map"])
def test_lgt_map_fit(make_weekly_data, seasonality, estimator):
    train_df, test_df, coef = make_weekly_data

    lgt = LGT(
        response_col="response",
        date_col="week",
        seasonality=seasonality,
        verbose=False,
        estimator=estimator,
    )

    lgt.fit(train_df)
    init_values = lgt._model.get_init_values()
    if seasonality:
        assert init_values["init_sea"].shape == (51,)
    else:
        assert init_values is None

    predict_df = lgt.predict(test_df)

    expected_num_parameters = len(lgt._model.get_model_param_names()) + 1
    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert len(lgt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("estimator", ["stan-mcmc", "pyro-svi"])
@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+", "+", "+", "+"],
        ["-", "-", "-", "-", "-", "-"],
        ["=", "=", "=", "=", "=", "="],
        ["+", "=", "+", "=", "-", "-"],
    ],
    ids=["positive_only", "negative_only", "regular_only", "mixed_signs"],
)
def test_lgt_full_with_regression(make_weekly_data, estimator, regressor_signs):
    train_df, test_df, coef = make_weekly_data

    if estimator == "stan-mcmc":
        lgt = LGT(
            response_col="response",
            date_col="week",
            regressor_col=train_df.columns.tolist()[2:],
            regressor_sign=regressor_signs,
            prediction_percentiles=[5, 95],
            seasonality=52,
            num_warmup=50,
            num_sample=50,
            verbose=False,
            estimator=estimator,
        )
    elif estimator == "pyro-svi":
        lgt = LGT(
            response_col="response",
            date_col="week",
            regressor_col=train_df.columns.tolist()[2:],
            regressor_sign=regressor_signs,
            prediction_percentiles=[5, 95],
            seasonality=52,
            num_steps=10,
            verbose=False,
            estimator=estimator,
        )
    else:
        return None

    lgt.fit(train_df)
    predict_df = lgt.predict(test_df)

    regression_out = lgt.get_regression_coefs()
    num_regressors = regression_out.shape[0]

    expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]
    expected_shape = (51, len(expected_columns))
    expected_regression_shape = (6, 7)

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert regression_out.shape == expected_regression_shape
    assert num_regressors == len(train_df.columns.tolist()[2:])


@pytest.mark.parametrize("estimator", ["stan-mcmc", "pyro-svi"])
@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+", "+", "+", "+"],
        ["-", "-", "-", "-", "-", "-"],
        ["=", "=", "=", "=", "=", "="],
        ["+", "=", "+", "=", "-", "-"],
    ],
    ids=["positive_only", "negative_only", "regular_only", "mixed_signs"],
)
@pytest.mark.parametrize("point_method", ["mean", "median"])
def test_lgt_aggregated_with_regression(
    make_weekly_data, estimator, regressor_signs, point_method
):
    train_df, test_df, coef = make_weekly_data

    if estimator == "stan-mcmc":
        lgt = LGT(
            response_col="response",
            date_col="week",
            regressor_col=train_df.columns.tolist()[2:],
            regressor_sign=regressor_signs,
            seasonality=52,
            num_warmup=50,
            num_sample=50,
            verbose=False,
            estimator=estimator,
        )
    elif estimator == "pyro-svi":
        lgt = LGT(
            response_col="response",
            date_col="week",
            regressor_col=train_df.columns.tolist()[2:],
            regressor_sign=regressor_signs,
            seasonality=52,
            num_steps=10,
            verbose=False,
            estimator=estimator,
        )
    else:
        return None

    lgt.fit(train_df, point_method=point_method)
    predict_df = lgt.predict(test_df)

    regression_out = lgt.get_regression_coefs()
    num_regressors = regression_out.shape[0]

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_regression_shape = (6, 7)

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert regression_out.shape == expected_regression_shape
    assert num_regressors == len(train_df.columns.tolist()[2:])

    predict_df = lgt.predict(test_df, decompose=True)
    assert any(predict_df["regression"].values)


@pytest.mark.parametrize("regressor_signs", [["=", "=", "+"]], ids=["positive_mixed"])
def test_lgt_mixed_signs_and_order(iclaims_training_data, regressor_signs):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])
    raw_regressor_col = ["trend.unemploy", "trend.filling", "trend.job"]
    new_regressor_col = [raw_regressor_col[idx] for idx in [2, 1, 0]]
    new_regressor_signs = [regressor_signs[idx] for idx in [2, 1, 0]]
    # mixing ordering of cols in df of prediction
    new_df = df[["claims", "week"] + new_regressor_col]

    lgt = LGT(
        response_col="claims",
        date_col="week",
        regressor_col=raw_regressor_col,
        regressor_sign=regressor_signs,
        estimator="stan-map",
        seasonality=52,
        seed=8888,
    )
    lgt.fit(df)
    predicted_df_v1 = lgt.predict(df)
    predicted_df_v2 = lgt.predict(new_df)

    # mixing ordering of signs
    lgt_new = LGT(
        response_col="claims",
        date_col="week",
        regressor_col=new_regressor_col,
        regressor_sign=new_regressor_signs,
        estimator="stan-map",
        seasonality=52,
        seed=8888,
    )
    lgt_new.fit(df)
    predicted_df_v3 = lgt_new.predict(df)
    predicted_df_v4 = lgt_new.predict(new_df)

    pred_v1 = predicted_df_v1["prediction"].values
    pred_v2 = predicted_df_v2["prediction"].values
    pred_v3 = predicted_df_v3["prediction"].values
    pred_v4 = predicted_df_v4["prediction"].values

    # they should be all identical; ordering of signs or columns in prediction show not matter
    assert np.allclose(pred_v1, pred_v2, atol=1e-2)
    assert np.allclose(pred_v1, pred_v3, atol=1e-2)
    assert np.allclose(pred_v1, pred_v4, atol=1e-2)


@pytest.mark.parametrize("prediction_percentiles", [None, [5, 10, 95]])
def test_lgt_prediction_percentiles(iclaims_training_data, prediction_percentiles):
    df = iclaims_training_data

    lgt = LGT(
        response_col="claims",
        date_col="week",
        seasonality=52,
        num_warmup=50,
        num_sample=50,
        seed=8888,
        prediction_percentiles=prediction_percentiles,
        estimator="stan-mcmc",
    )

    if not prediction_percentiles:
        p_labels = ["_5", "", "_95"]
    else:
        p_labels = ["_5", "_10", "", "_95"]

    lgt.fit(df)
    predicted_df = lgt.predict(df)
    expected_columns = ["week"] + ["prediction" + p for p in p_labels]
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]

    predicted_df = lgt.predict(df, decompose=True)
    predicted_components = [
        "prediction",
        PredictionKeys.TREND.value,
        PredictionKeys.SEASONALITY.value,
        PredictionKeys.REGRESSION.value,
    ]

    expected_columns = ["week"]
    for pc in predicted_components:
        for p in p_labels:
            expected_columns.append(pc + p)
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]


@pytest.mark.parametrize("estimator", ["stan-mcmc"])
@pytest.mark.parametrize(
    "regressor_signs",
    [
        ["+", "+", "+", "+", "+", "+"],
        ["=", "=", "=", "=", "=", "="],
        ["+", "=", "+", "=", "+", "+"],
    ],
    ids=["positive_only", "regular_only", "mixed_signs"],
)
@pytest.mark.parametrize("seasonality", [1, 52])
def test_lgt_full_reproducibility(
    make_weekly_data, estimator, regressor_signs, seasonality
):
    train_df, test_df, coef = make_weekly_data

    lgt_first = LGT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        regressor_sign=regressor_signs,
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        num_warmup=50,
        num_sample=50,
        verbose=False,
        estimator=estimator,
    )

    # first fit and predict
    lgt_first.fit(train_df)
    posteriors_first = copy(lgt_first._posterior_samples)
    predict_df_first = lgt_first.predict(test_df)
    regression_out_first = lgt_first.get_regression_coefs()

    # second fit and predict
    # note a new instance must be created to reset the seed
    # note both fit and predict contain random generation processes
    lgt_second = LGT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        regressor_sign=regressor_signs,
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        num_warmup=50,
        num_sample=50,
        verbose=False,
        estimator=estimator,
    )

    lgt_second.fit(train_df)
    posteriors_second = copy(lgt_second._posterior_samples)
    predict_df_second = lgt_second.predict(test_df)
    regression_out_second = lgt_second.get_regression_coefs()

    # assert same posterior keys
    assert set(posteriors_first.keys()) == set(posteriors_second.keys())

    # assert posterior draws are reproducible
    for k, v in posteriors_first.items():
        assert np.allclose(posteriors_first[k], posteriors_second[k])

    # assert identical regression columns
    # this is also checked in posterior samples, but an extra layer just in case
    # since this one very commonly retrieved by end users
    assert regression_out_first.equals(regression_out_second)

    # assert prediction is reproducible
    assert predict_df_first.equals(predict_df_second)


@pytest.mark.parametrize("seasonality", [1, 52])
def test_lgt_map_reproducibility(make_weekly_data, seasonality):
    train_df, test_df, coef = make_weekly_data

    lgt1 = LGT(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        estimator="stan-map",
    )

    # first fit and predict
    lgt1.fit(train_df)
    posteriors1 = copy(lgt1._point_posteriors["map"])
    prediction1 = lgt1.predict(test_df)

    # second fit and predict
    # note a new instance must be created to reset the seed
    # note both fit and predict contain random generation processes
    lgt2 = LGT(
        response_col="response",
        date_col="week",
        prediction_percentiles=[5, 95],
        seasonality=seasonality,
        estimator="stan-map",
    )

    lgt2.fit(train_df)
    posteriors2 = copy(lgt2._point_posteriors["map"])
    prediction2 = lgt2.predict(test_df)

    # assert same posterior keys
    assert set(posteriors1.keys()) == set(posteriors2.keys())

    # assert posterior draws are reproducible
    for k, v in posteriors1.items():
        assert np.allclose(posteriors1[k], posteriors2[k])

    # assert prediction is reproducible
    assert np.allclose(
        prediction1["prediction"].values, prediction2["prediction"].values
    )


@pytest.mark.parametrize("level_sm_input", [0.0001, 0.5, 1.0])
@pytest.mark.parametrize("seasonality_sm_input", [0.0, 0.5, 1.0])
@pytest.mark.parametrize("slope_sm_input", [0.0, 0.5, 1.0])
def test_lgt_fixed_sm_input(
    make_weekly_data, level_sm_input, seasonality_sm_input, slope_sm_input
):
    train_df, test_df, coef = make_weekly_data

    lgt = LGT(
        response_col="response",
        date_col="week",
        regressor_col=train_df.columns.tolist()[2:],
        level_sm_input=level_sm_input,
        seasonality_sm_input=seasonality_sm_input,
        slope_sm_input=slope_sm_input,
        estimator="stan-map",
        seasonality=52,
        verbose=False,
    )

    lgt.fit(train_df)
    predict_df = lgt.predict(test_df)

    regression_out = lgt.get_regression_coefs()
    num_regressors = regression_out.shape[0]

    expected_columns = ["week", "prediction"]
    expected_shape = (51, len(expected_columns))
    expected_regression_shape = (6, 3)

    assert predict_df.shape == expected_shape
    assert predict_df.columns.tolist() == expected_columns
    assert regression_out.shape == expected_regression_shape
    assert num_regressors == len(train_df.columns.tolist()[2:])


@pytest.mark.parametrize("estimator", ["stan-mcmc", "stan-map"])
def test_lgt_missing(iclaims_training_data, estimator):
    df = iclaims_training_data
    missing_idx = np.array([10, 20, 30, 40, 41, 42, 43, 44, df.shape[0] - 1])
    df.loc[missing_idx, "claims"] = np.nan

    dlt = LGT(
        response_col="claims",
        date_col="week",
        seasonality=52,
        verbose=False,
        estimator=estimator,
    )

    dlt.fit(df)
    predicted_df = dlt.predict(df)
    if estimator == "stan-map":
        expected_columns = ["week", "prediction"]
    elif estimator == "stan-mcmc":
        expected_columns = ["week", "prediction_5", "prediction", "prediction_95"]

    assert all(~np.isnan(predicted_df["prediction"]))
    assert predicted_df.columns.tolist() == expected_columns
    assert predicted_df.shape[0] == df.shape[0]


def test_lgt_map_single_regressor(iclaims_training_data):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])
    regressor_col = ["trend.unemploy"]

    lgt = LGT(
        response_col="claims",
        date_col="week",
        regressor_col=regressor_col,
        estimator="stan-map",
        seasonality=52,
        seed=8888,
    )
    lgt.fit(df)
    predicted_df = lgt.predict(df)

    expected_num_parameters = len(lgt._model.get_model_param_names()) + 1
    expected_columns = ["week", "prediction"]

    assert predicted_df.shape[0] == df.shape[0]
    assert predicted_df.columns.tolist() == expected_columns
    assert len(lgt._posterior_samples) == expected_num_parameters


@pytest.mark.parametrize("estimator", ["stan-mcmc", "pyro-svi"])
@pytest.mark.parametrize("keep_samples", [True, False])
@pytest.mark.parametrize("point_method", ["mean", "median"])
def test_lgt_is_fitted(iclaims_training_data, estimator, keep_samples, point_method):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])
    regressor_col = ["trend.unemploy"]

    if estimator == "stan-mcmc":
        lgt = LGT(
            response_col="claims",
            date_col="week",
            regressor_col=regressor_col,
            seasonality=52,
            seed=8888,
            num_warmup=50,
            num_sample=50,
            verbose=False,
            estimator=estimator,
        )
    elif estimator == "pyro-svi":
        lgt = LGT(
            response_col="claims",
            date_col="week",
            regressor_col=regressor_col,
            seasonality=52,
            seed=8888,
            num_steps=10,
            verbose=False,
            estimator=estimator,
        )
    lgt.fit(df, keep_samples=keep_samples, point_method=point_method)
    is_fitted = lgt.is_fitted()

    # still True when keep_samples is False
    assert is_fitted


@pytest.mark.parametrize("estimator", ["stan-mcmc", "stan-map", "pyro-svi"])
@pytest.mark.parametrize("random_seed", [10, 100])
def test_lgt_predict_seed(make_weekly_data, estimator, random_seed):
    train_df, test_df, coef = make_weekly_data
    args = {
        "response_col": "response",
        "date_col": "week",
        "seasonality": 52,
        "n_bootstrap_draws": 100,
        "verbose": False,
        "estimator": estimator,
    }

    if estimator == "stan-mcmc":
        args.update({"num_warmup": 50, "num_sample": 100})
    elif estimator == "pyro-svi":
        args.update({"num_steps": 10})

    lgt = LGT(**args)
    lgt.fit(train_df)
    predict_df1 = lgt.predict(test_df, seed=random_seed)
    predict_df2 = lgt.predict(test_df, seed=random_seed)

    assert all(predict_df1["prediction"].values == predict_df2["prediction"].values)


@pytest.mark.parametrize(
    "idx_range",
    [
        [0, 100],
        [0, 50],
        [50, 100],
        [50, 80],
        [50, 120],
        [100, 150],
    ],
    ids=[
        "train-start-to-train-end",
        "train-start-to-middle",
        "middle-to-train-end",
        "train-period-subset",
        "train-test-period-cross",
        "completely-test-period",
    ],
)
def test_lgt_predict_range(make_weekly_data, idx_range):
    train_cut_off = 100
    base_df, _, _ = make_weekly_data
    train_df = base_df[:train_cut_off].reset_index(drop=True)
    predict_df = base_df[idx_range[0] : idx_range[1]].reset_index(drop=True)

    lgt = LGT(
        response_col="response",
        date_col="week",
        seasonality=52,
        verbose=False,
        estimator="stan-map",
    )
    lgt.fit(train_df)
    lgt.predict(predict_df)



import pandas as pd
import inspect
from tqdm.auto import tqdm
from itertools import product
from collections.abc import Mapping, Iterable

from ..diagnostics.metrics import smape
from ..diagnostics.backtest import BackTester
from ..exceptions import IllegalArgument
from ..forecaster import MAPForecaster

import logging

logger = logging.getLogger("orbit")


def grid_search_orbit(
    param_grid,
    model,
    df,
    eval_method="backtest",
    min_train_len=None,
    incremental_len=None,
    forecast_len=None,
    n_splits=None,
    metrics=None,
    criteria="min",
    verbose=False,
    **kwargs,
):
    """A gird search utility to tune the hyperparameters for orbit template using the orbit.diagnostics.backtest modules.
    Parameters
    ----------
    param_grid : dict
        a dict with candidate values for hyper-params to be tuned
    model : object
        model object
    df : pd.DataFrame
    eval_method : str
        "backtest" or "bic"
    min_train_len : int
        scheduling parameter in backtest
    incremental_len : int
        scheduling parameter in backtest
    forecast_len : int
        scheduling parameter in backtest
    n_splits : int
        scheduling parameter in backtest
    metrics : callable
        metric function in use when evel_method is "backtest";
        if not provided, default will be set as smape defined in orbit.diagnostics.metrics
    criteria : str
        "min" or "max"
    verbose : bool

    Return
    ------
    dict:
        best hyperparams
    pd.DataFrame:
        data frame of tuning results

    """
    # def _get_params(model):
    #     # get all the model params for orbit typed template
    #     params = {}
    #     for key, val in model.__dict__.items():
    #         if not key.startswith('_') and key != 'estimator':
    #             params[key] = val

    #     for key, val in model.__dict__['estimator'].__dict__.items():
    #         if not key.startswith('_') and key != 'stan_init':
    #             params[key] = val

    #     return params.copy()
    if eval_method not in ["backtest", "bic"]:
        raise IllegalArgument(
            "Invalid input of eval_method. Argument not in ['backtest', 'bic']"
        )

    if eval_method == "bic" and criteria != "min":
        logger.info("crtieria is enforced to be min when using 'bic' as eval_method.")
        criteria = "min"

    if criteria not in ["min", "max"]:
        raise IllegalArgument(
            "Invalid input of criteria. Argument not in ['min', 'max']"
        )

    def _get_params(model):
        init_args_tmpl = dict()
        init_args = dict()

        # get all the signatures in the hierarchy of model templates
        for cls in inspect.getmro(model._model.__class__):
            sig = inspect.signature(cls)
            for key in sig.parameters.keys():
                if key != "kwargs":
                    if hasattr(model._model, key):
                        init_args_tmpl[key] = getattr(model._model, key)
        # get all the signatures in the hierarchy of forecaster
        for cls in inspect.getmro(model.__class__):
            sig = inspect.signature(cls)
            for key in sig.parameters.keys():
                if key != "kwargs":
                    if hasattr(model, key):
                        init_args[key] = getattr(model, key)
        # deal with the estimator separately
        for cls in inspect.getmro(model.estimator_type):
            sig = inspect.signature(cls)
            for key in sig.parameters.keys():
                if key != "kwargs":
                    if hasattr(model.estimator, key):
                        init_args[key] = getattr(model.estimator, key)

        return init_args_tmpl.copy(), init_args.copy()

    param_list_dict = generate_param_args_list(param_grid)
    params_tmpl, params = _get_params(model)
    res = pd.DataFrame(param_list_dict)
    metric_values = list()

    for tuned_param_dict in tqdm(param_list_dict):
        if verbose:
            logger.info("tuning hyper-params {}".format(tuned_param_dict))

        params_ = params.copy()
        params_tmpl_ = params_tmpl.copy()
        for key, val in tuned_param_dict.items():
            if key in params_tmpl_.keys():
                params_tmpl_[key] = val
            elif key in params_.keys():
                params_[key] = val
            else:
                raise IllegalArgument(
                    "tuned hyper-param {} is not in the model's parameters".format(key)
                )

        # it is safer to re-instantiate a model object than using deepcopy...
        new_model_template = model._model.__class__(**params_tmpl_)
        new_model = model.__class__(model=new_model_template, **params_)

        if eval_method == "backtest":
            bt = BackTester(
                model=new_model,
                df=df,
                min_train_len=min_train_len,
                n_splits=n_splits,
                incremental_len=incremental_len,
                forecast_len=forecast_len,
                **kwargs,
            )
            bt.fit_predict()
            # TODO: should we assert len(metrics) == 1?
            if metrics is None:
                metrics = smape
            metric_val = bt.score(metrics=[metrics]).metric_values[0]
        elif eval_method == "bic":
            if isinstance(new_model, MAPForecaster):
                # pass thru arg for estimator
                new_model.fit(df)
                metric_val = new_model.get_bic()
            else:
                raise IllegalArgument(
                    "eval_method 'bic' only supports 'stan-map' estimator for now."
                )
        if verbose:
            logger.info("tuning metric:{:-.5g}".format(metric_val))
        metric_values.append(metric_val)

    res["metrics"] = metric_values

    best_params = (
        res[res["metrics"] == res["metrics"].apply(criteria)]
        .drop("metrics", axis=1)
        .to_dict("records")
    )

    return best_params, res


def generate_param_args_list(param_grid):
    """An utils similar to sci-kit learn package to generate combinations of args based on spaces of parameters
    provided from users in a dictionary of list format.


    Parameters
    ----------
    param_grid: dict of list
    dictionary where key represents the arg and value represents the list of values proposed for the grid

    Returns
    -------
    list of dict
    the iterated products of all combinations of params generated based on the input param grid
    """

    # an internal function to mimic the ParameterGrid from scikit-learn

    def _yield_param_grid(param_grid):
        if not isinstance(param_grid, (Mapping, Iterable)):
            raise TypeError(
                "Parameter grid is not a dict or a list ({!r})".format(param_grid)
            )

        if isinstance(param_grid, Mapping):
            # wrap dictionary in a singleton list to support either dict
            # or list of dicts
            param_grid = [param_grid]

        for p in param_grid:
            # Always sort the keys of a dictionary, for reproducibility
            items = sorted(p.items())
            if not items:
                yield {}
            else:
                keys, values = zip(*items)
                for v in product(*values):
                    params = dict(zip(keys, v))
                    yield params

    return list(_yield_param_grid(param_grid))



# the following lines are added to fix unit test error
# or else the following line will give the following error
# TclError: no display name and no $DISPLAY environment variable
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
import math
import os
import pkg_resources
import statsmodels.api as sm
from scipy import stats

from ..constants.constants import PredictionKeys
from orbit.utils.general import is_empty_dataframe, is_ordered_datetime
from ..constants.constants import BacktestFitKeys
from ..constants.palette import PredictionPaletteClassic as PredPal
from orbit.constants import palette
from orbit.diagnostics.metrics import smape
from orbit.utils.plot import orbit_style_decorator
from ..exceptions import PlotException


import logging

logger = logging.getLogger("orbit")


@orbit_style_decorator
def plot_predicted_data(
    training_actual_df,
    predicted_df,
    date_col,
    actual_col,
    pred_col=PredictionKeys.PREDICTION.value,
    prediction_percentiles=None,
    title="",
    test_actual_df=None,
    is_visible=True,
    figsize=None,
    path=None,
    fontsize=None,
    line_plot=False,
    markersize=50,
    lw=2,
    linestyle="-",
):
    """plot training actual response together with predicted data; if actual response of predicted
    data is there, plot it too.

    Parameters
    ----------
    training_actual_df : pd.DataFrame
        training actual response data frame. two columns required: actual_col and date_col
    predicted_df : pd.DataFrame
        predicted data response data frame. two columns required: actual_col and pred_col. If
        user provide prediction_percentiles, it needs to include them as well in such
        `prediction_{x}` where x is the correspondent percentiles
    prediction_percentiles : list
        list of two elements indicates the lower and upper percentiles
    date_col : str
        the date column name
    actual_col : str
    pred_col : str
    title : str
        title of the plot
    test_actual_df : pd.DataFrame
       test actual response dataframe. two columns required: actual_col and date_col
    is_visible : boolean
        whether we want to show the plot. If called from unittest, is_visible might = False.
    figsize : tuple
        figsize pass through to `matplotlib.pyplot.figure()`
    path : str
        path to save the figure
    fontsize : int; optional
        fontsize of the title
    line_plot : bool; default False
        if True, make line plot for observations; otherwise, make scatter plot for observations
    markersize : int; optional
        point marker size
    lw : int; optional
        out-of-sample prediction line width
    linestyle : str
        linestyle of prediction plot

    Returns
    -------
        matplotlib axes object
    """

    if is_empty_dataframe(training_actual_df) or is_empty_dataframe(predicted_df):
        raise ValueError("No prediction data or training response to plot.")

    if not is_ordered_datetime(predicted_df[date_col]):
        raise ValueError("Prediction df dates is not ordered.")

    plot_confid = False
    if prediction_percentiles is None:
        _pred_percentiles = [5, 95]
    else:
        _pred_percentiles = prediction_percentiles

    if len(_pred_percentiles) != 2:
        raise ValueError(
            "prediction_percentiles has to be None or a list with length=2."
        )

    confid_cols = [
        "prediction_{}".format(_pred_percentiles[0]),
        "prediction_{}".format(_pred_percentiles[1]),
    ]

    if set(confid_cols).issubset(predicted_df.columns):
        plot_confid = True

    if not figsize:
        figsize = (16, 8)

    if not fontsize:
        fontsize = 16

    _training_actual_df = training_actual_df.copy()
    _predicted_df = predicted_df.copy()
    _training_actual_df[date_col] = pd.to_datetime(_training_actual_df[date_col])
    _predicted_df[date_col] = pd.to_datetime(_predicted_df[date_col])

    fig, ax = plt.subplots(facecolor="w", figsize=figsize)

    if line_plot:
        ax.plot(
            _training_actual_df[date_col].values,
            _training_actual_df[actual_col].values,
            marker=None,
            color=PredPal.ACTUAL_OBS.value,
            lw=lw,
            label="train response",
            linestyle=linestyle,
        )
    else:
        ax.scatter(
            _training_actual_df[date_col].values,
            _training_actual_df[actual_col].values,
            marker=".",
            color=PredPal.ACTUAL_OBS.value,
            alpha=0.8,
            s=markersize,
            label="train response",
        )

    ax.plot(
        _predicted_df[date_col].values,
        _predicted_df[pred_col].values,
        marker=None,
        color=PredPal.PREDICTION_LINE.value,
        lw=lw,
        label=PredictionKeys.PREDICTION.value,
        linestyle=linestyle,
    )

    # vertical line separate training and prediction
    if _training_actual_df[date_col].values[-1] < _predicted_df[date_col].values[-1]:
        ax.axvline(
            x=_training_actual_df[date_col].values[-1],
            color=PredPal.HOLDOUT_VERTICAL_LINE.value,
            alpha=0.5,
            linestyle="--",
        )

    if test_actual_df is not None:
        test_actual_df = test_actual_df.copy()
        test_actual_df[date_col] = pd.to_datetime(test_actual_df[date_col])
        if line_plot:
            ax.plot(
                test_actual_df[date_col].values,
                test_actual_df[actual_col].values,
                marker=None,
                color=PredPal.TEST_OBS.value,
                lw=lw,
                label="train response",
                linestyle=linestyle,
            )
        else:
            ax.scatter(
                test_actual_df[date_col].values,
                test_actual_df[actual_col].values,
                marker=".",
                color=PredPal.TEST_OBS.value,
                s=markersize,
                label="test response",
            )

    # prediction intervals
    if plot_confid:
        ax.fill_between(
            _predicted_df[date_col].values,
            _predicted_df[confid_cols[0]],
            _predicted_df[confid_cols[1]],
            facecolor=PredPal.PREDICTION_INTERVAL.value,
            alpha=0.3,
        )

    ax.set_title(title, fontsize=fontsize)
    # ax.grid(True, which='major', c='gray', ls='-', lw=1, alpha=0.5) --comment out since we have orbit style
    ax.legend()
    if path:
        fig.savefig(path)
    if is_visible:
        plt.show()
    else:
        plt.close()

    return ax


@orbit_style_decorator
def plot_predicted_components(
    predicted_df,
    date_col,
    prediction_percentiles=None,
    plot_components=None,
    title="",
    figsize=None,
    path=None,
    fontsize=None,
    is_visible=True,
):
    """Plot predicted components with the data frame of decomposed prediction where components
     has been pre-defined as `trend`, `seasonality` and `regression`.

     Parameters
     ----------
     predicted_df : pd.DataFrame
         predicted data response data frame. two columns required: actual_col and pred_col. If
         user provide pred_percentiles_col, it needs to include them as well.
     date_col : str
         the date column name
     prediction_percentiles : list
         a list should consist exact two elements which will be used to plot as lower and upper bound of
         confidence interval
     plot_components : list
         a list of strings to show the label of components to be plotted; by default, it uses values in
         `orbit.constants.constants.PredictedComponents`.
     title : str; optional
         title of the plot
     figsize : tuple; optional
         figsize pass through to `matplotlib.pyplot.figure()`
     path : str; optional
         path to save the figure
     fontsize : int; optional
         fontsize of the title
     is_visible : boolean
         whether we want to show the plot. If called from unittest, is_visible might = False.

    Returns
    -------
         matplotlib axes object
    """

    _predicted_df = predicted_df.copy()
    _predicted_df[date_col] = pd.to_datetime(_predicted_df[date_col])
    if plot_components is None:
        plot_components = [
            PredictionKeys.TREND.value,
            PredictionKeys.SEASONALITY.value,
            PredictionKeys.REGRESSION.value,
        ]

    plot_components = [
        p for p in plot_components if p in _predicted_df.columns.tolist()
    ]
    nrows = len(plot_components)
    if not figsize:
        figsize = (16, 8)

    if not fontsize:
        fontsize = 16

    if prediction_percentiles is None:
        _pred_percentiles = [5, 95]
    else:
        _pred_percentiles = prediction_percentiles

    if len(_pred_percentiles) != 2:
        raise ValueError(
            "prediction_percentiles has to be None or a list with length=2."
        )

    fig, axes = plt.subplots(nrows=nrows, ncols=1, figsize=figsize, squeeze=False)
    axes = axes.flatten()
    for ax, comp in zip(axes, plot_components):
        y = predicted_df[comp].values
        ax.plot(
            _predicted_df[date_col],
            y,
            marker=None,
            color=PredPal.PREDICTION_INTERVAL.value,
        )
        confid_cols = [
            "{}_{}".format(comp, _pred_percentiles[0]),
            "{}_{}".format(comp, _pred_percentiles[1]),
        ]
        if set(confid_cols).issubset(predicted_df.columns):
            ax.fill_between(
                _predicted_df[date_col].values,
                _predicted_df[confid_cols[0]],
                _predicted_df[confid_cols[1]],
                facecolor=PredPal.PREDICTION_INTERVAL.value,
                alpha=0.3,
            )
        ax.set_title(comp, fontsize=fontsize)

    plt.suptitle(title, fontsize=fontsize)
    fig.tight_layout()

    if path:
        plt.savefig(path)
    if is_visible:
        plt.show()
    else:
        plt.close()

    return axes


@orbit_style_decorator
def plot_bt_predictions(
    bt_pred_df,
    metrics=smape,
    split_key_list=None,
    ncol=2,
    figsize=None,
    include_vline=True,
    title="",
    fontsize=20,
    path=None,
    is_visible=True,
):
    """function to plot and visualize the prediction results from back testing.

    bt_pred_df : data frame
        the output of `orbit.diagnostics.backtest.BackTester.fit_predict()`, which includes the actuals/predictions
        for all the splits
    metrics : callable
        the metric function
    split_key_list: list; default None
        with given model, which split keys to plot. If None, all the splits will be plotted
    ncol : int
        number of columns of the panel; number of rows will be decided accordingly
    figsize : tuple
        figure size
    include_vline : bool
        if plotting the vertical line to cut the in-sample and out-of-sample predictions for each split
    title : str
        title of the plot
    fontsize: int; optional
        fontsize of the title
    path : string
        path to save the figure
    is_visible : bool
        if displaying the figure
    """
    if figsize is None:
        figsize = (16, 8)

    metric_vals = bt_pred_df.groupby(BacktestFitKeys.SPLIT_KEY.value).apply(
        lambda x: metrics(
            x[~x[BacktestFitKeys.TRAIN_FLAG.value]][BacktestFitKeys.ACTUAL.value],
            x[~x[BacktestFitKeys.TRAIN_FLAG.value]][BacktestFitKeys.PREDICTED.value],
        )
    )

    if split_key_list is None:
        split_key_list_ = bt_pred_df[BacktestFitKeys.SPLIT_KEY.value].unique()
    else:
        split_key_list_ = split_key_list

    num_splits = len(split_key_list_)
    nrow = math.ceil(num_splits / ncol)
    fig, axes = plt.subplots(
        nrow,
        ncol,
        figsize=figsize,
        squeeze=False,
        facecolor="w",
        constrained_layout=False,
    )

    for idx, split_key in enumerate(split_key_list_):
        row_idx = idx // ncol
        col_idx = idx % ncol
        tmp = bt_pred_df[
            bt_pred_df[BacktestFitKeys.SPLIT_KEY.value] == split_key
        ].copy()
        axes[row_idx, col_idx].plot(
            tmp[BacktestFitKeys.DATE.value],
            tmp[BacktestFitKeys.PREDICTED.value],
            # linewidth=2,
            color=PredPal.PREDICTION_LINE.value,
        )
        axes[row_idx, col_idx].scatter(
            tmp[BacktestFitKeys.DATE.value],
            tmp[BacktestFitKeys.ACTUAL.value],
            label=BacktestFitKeys.ACTUAL.value,
            color=PredPal.ACTUAL_OBS.value,
            alpha=0.6,
            s=8,
        )
        # axes[row_idx, col_idx].grid(True, which='major', c='gray', ls='-', lw=1, alpha=0.4)

        axes[row_idx, col_idx].set_title(
            label="split {}; {} {:.3f}".format(
                split_key, metrics.__name__, metric_vals[split_key]
            )
        )
        if include_vline:
            cutoff_date = tmp[~tmp[BacktestFitKeys.TRAIN_FLAG.value]][
                BacktestFitKeys.DATE.value
            ].min()
            axes[row_idx, col_idx].axvline(
                x=cutoff_date,
                linestyle="--",
                color=PredPal.HOLDOUT_VERTICAL_LINE.value,
                # linewidth=4,
                alpha=0.8,
            )

    plt.suptitle(title, fontsize=fontsize)
    fig.tight_layout()
    if path:
        fig.savefig(path)
    if is_visible:
        plt.show()
    else:
        plt.close()

    return axes


@orbit_style_decorator
def plot_bt_predictions2(
    bt_pred_df,
    metrics=smape,
    split_key_list=None,
    figsize=None,
    include_vline=True,
    title="",
    fontsize=20,
    markersize=50,
    lw=2,
    fig_dir=None,
    is_visible=True,
    fix_xylim=True,
    export_gif=False,
):
    """a different style backtest plot compare to `plot_bt_prediction` where it writes separate plot for each split;
    this is also used to produce an animation to summarize every split
    """
    if figsize is None:
        figsize = (16, 8)

    if fig_dir:
        if not os.path.isdir(fig_dir) or not os.path.exists(fig_dir):
            raise PlotException(
                "Invalid or non-existing directory use specified: {}.".format(
                    os.path.abspath(fig_dir)
                )
            )
        fig_paths = list()

    metric_vals = bt_pred_df.groupby(BacktestFitKeys.SPLIT_KEY.value).apply(
        lambda x: metrics(
            x[~x[BacktestFitKeys.TRAIN_FLAG.value]][BacktestFitKeys.ACTUAL.value],
            x[~x[BacktestFitKeys.TRAIN_FLAG.value]][BacktestFitKeys.PREDICTED.value],
        )
    )

    if split_key_list is None:
        split_key_list_ = bt_pred_df[BacktestFitKeys.SPLIT_KEY.value].unique()
    else:
        split_key_list_ = split_key_list

    if fix_xylim:
        all_values = np.concatenate(
            (
                bt_pred_df[BacktestFitKeys.ACTUAL.value].values,
                bt_pred_df[BacktestFitKeys.PREDICTED.value].values,
            )
        )
        ylim = (np.min(all_values) * 0.99, np.max(all_values) * 1.01)
        xlim = (
            bt_pred_df[BacktestFitKeys.DATE.value].values[0],
            bt_pred_df[BacktestFitKeys.DATE.value].values[-1],
        )

    for idx, split_key in enumerate(split_key_list_):
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        tmp = bt_pred_df[
            bt_pred_df[BacktestFitKeys.SPLIT_KEY.value] == split_key
        ].copy()
        ax.plot(
            tmp[BacktestFitKeys.DATE.value],
            tmp[BacktestFitKeys.PREDICTED.value],
            color=PredPal.PREDICTION_LINE.value,
            lw=lw,
        )

        train_df = tmp.loc[tmp[BacktestFitKeys.TRAIN_FLAG.value], :]
        ax.scatter(
            train_df[BacktestFitKeys.DATE.value],
            train_df[BacktestFitKeys.ACTUAL.value],
            marker=".",
            color=PredPal.ACTUAL_OBS.value,
            alpha=0.8,
            s=markersize,
            label="train response",
        )

        test_df = tmp.loc[~tmp[BacktestFitKeys.TRAIN_FLAG.value], :]
        ax.scatter(
            test_df[BacktestFitKeys.DATE.value],
            test_df[BacktestFitKeys.ACTUAL.value],
            marker=".",
            color=PredPal.TEST_OBS.value,
            alpha=0.8,
            s=markersize,
            label="test response",
        )

        ax.set_title(
            label="split {}; {} {:.3f}".format(
                split_key, metrics.__name__, metric_vals[split_key]
            )
        )
        if include_vline:
            cutoff_date = tmp[~tmp[BacktestFitKeys.TRAIN_FLAG.value]][
                BacktestFitKeys.DATE.value
            ].min()
            ax.axvline(
                x=cutoff_date,
                linestyle="--",
                color=PredPal.HOLDOUT_VERTICAL_LINE.value,
                alpha=0.8,
            )
        if fix_xylim:
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

        ax.legend()
        plt.suptitle(title, fontsize=fontsize)
        fig.tight_layout()

        if fig_dir:
            fig_path = "{}/splits_{}.png".format(fig_dir, idx)
            fig_paths.append(fig_path)
            fig.savefig(fig_path)
        if is_visible:
            plt.show()
        else:
            plt.close()

    if fig_dir and export_gif:
        package_name = "imageio"
        try:
            pkg_resources.get_distribution(package_name)
            import imageio

            with imageio.get_writer(
                "{}/orbit-backtest.gif".format(fig_dir), mode="I"
            ) as writer:
                for fig_path in fig_paths:
                    image = imageio.imread(fig_path)
                    writer.append_data(image)
        except pkg_resources.DistributionNotFound:
            logger.error(
                (
                    "{} not installed, which is necessary for gif animation".format(
                        package_name
                    )
                )
            )


# TODO: update palette
@orbit_style_decorator
def metric_horizon_barplot(
    df,
    model_col="model",
    pred_horizon_col="pred_horizon",
    metric_col="smape",
    bar_width=0.1,
    path=None,
    figsize=None,
    fontsize=None,
    is_visible=False,
):
    if not figsize:
        figsize = [20, 6]

    if not fontsize:
        fontsize = 10

    plt.rcParams["figure.figsize"] = figsize
    models = df[model_col].unique()
    metric_horizons = df[pred_horizon_col].unique()
    n_models = len(models)
    palette = sns.color_palette("colorblind", n_models)

    # set height of bar
    bars = list()
    for m in models:
        bars.append(list(df[df[model_col] == m][metric_col]))

    # set position of bar on X axis
    r = list()
    r.append(np.arange(len(bars[0])))
    for idx in range(n_models - 1):
        r.append([x + bar_width for x in r[idx]])

    # make the plot
    for idx in range(n_models):
        plt.bar(
            r[idx],
            bars[idx],
            color=palette[idx],
            width=bar_width,
            edgecolor="white",
            label=models[idx],
        )

    # add xticks on the middle of the group bars
    plt.xlabel("predict-horizon", fontweight="bold")
    plt.xticks([x + bar_width for x in range(len(bars[0]))], metric_horizons)

    # create legend & show graphic
    plt.legend()
    plt.title("Model Comparison with {}".format(metric_col), fontsize=fontsize)

    if path:
        plt.savefig(path)

    if is_visible:
        plt.show()
    else:
        plt.close()


@orbit_style_decorator
def params_comparison_boxplot(
    data,
    var_names,
    model_names,
    color_list=sns.color_palette(),
    title="Params Comparison",
    fig_size=(10, 6),
    box_width=0.1,
    box_distance=0.2,
    showfliers=False,
):
    """compare the distribution of parameters from different models uisng a boxplot.
    Parameters:
        data : a list of dict with keys as the parameters of interest
        var_names : a list of strings, the labels of the parameters to compare
        model_names : a list of strings, the names of models to compare
        color_list : a list of strings, the color to use for differentiating models
        title : string
            the title of the chart
        fig_size : tuple
            figure size
        box_width : float
            width of the boxes in the boxplot
        box_distance : float
            the distance between each boxes in the boxplot
        showfliers  : boolean
            show outliers in the chart if set as True

    Returns:
        a boxplot comparing parameter distributions from different models side by side
    """

    fig, ax = plt.subplots(1, 1, figsize=fig_size)
    handles = []
    n_models = len(model_names)
    pos = []

    if n_models % 2 == 0:
        for n in range(1, int(n_models / 2) + 1):
            pos.append(round(box_distance * (-1) ** (n_models - 1) * n, 1))
            pos.append(round(box_distance * (-1) ** (n_models) * n, 1))

    else:
        for n in range(1, int((n_models - 1) / 2) + 1):
            pos.append(0)
            pos.append(round(box_distance * (-1) ** (n_models - 1) * n, 1))
            pos.append(round(box_distance * (-1) ** (n_models) * n, 1))

    pos = sorted(pos)

    for i in range(len(model_names)):
        plt_arr = []
        for var in var_names:
            plt_arr.append(data[i][var].flatten())
        plt_arr = np.vstack(plt_arr).T
        globals()[f"bp{i}"] = ax.boxplot(
            plt_arr,
            positions=np.arange(plt_arr.shape[1]) + pos[i],
            widths=box_width,
            patch_artist=True,
            manage_ticks=False,
            boxprops=dict(facecolor=color_list[i]),
            medianprops=dict(color="black"),
            showfliers=showfliers,
        )
        handles.append(globals()[f"bp{i}"]["boxes"][0])

    plt.xticks(np.arange(len(var_names)), var_names)
    ax.legend(handles, model_names)
    plt.xlabel("params")
    plt.ylabel("value")
    plt.title(title)

    return ax


@orbit_style_decorator
def residual_diagnostic_plot(
    df,
    dist="norm",
    date_col="week",
    residual_col="residual",
    fitted_col="prediction",
    sparams=None,
):
    """
    Parameters
    ----------

    df : pd.DataFrame
    dist : str
    date_col : str
        column name of date
    residual_col : str
        column name of residual
    fitted_col: str
        column name of fitted value from model
    sparams : float or list
        extra parameters used in distribution such as t-dist

    Notes
    -----
    1. residual by time
    2. residual vs fitted
    3. residual histogram with vertical line as mean
    4. residuals qq plot
    5. residual ACF
    6. residual PACF
    """
    fig, ax = plt.subplots(3, 2, figsize=(15, 12))

    # plot 1 residual by time
    sns.lineplot(
        x=date_col,
        y=residual_col,
        data=df,
        ax=ax[0, 0],
        color=palette.OrbitPalette.BLUE.value,
        alpha=0.8,
        label="residual",
    )
    ax[0, 0].set_title("Residual by Time")
    ax[0, 0].legend()

    # plot 2 residual vs fitted
    sns.scatterplot(
        x=fitted_col,
        y=residual_col,
        data=df,
        ax=ax[0, 1],
        color=palette.OrbitPalette.BLUE.value,
        alpha=0.8,
        label="residual",
    )
    ax[0, 1].axhline(
        y=0,
        linestyle="--",
        color=palette.OrbitPalette.BLACK.value,
        alpha=0.5,
        label="0",
    )
    ax[0, 1].set_title("Residual vs Fitted")
    ax[0, 1].set_xlabel("fitted")
    ax[0, 1].legend()

    # plot 3 residual histogram with vertical line as mean
    sns.histplot(
        df[residual_col].values,
        kde=True,
        ax=ax[1, 0],
        color=palette.OrbitPalette.BLUE.value,
        label="residual",
        edgecolor="white",
        alpha=0.5,
        facecolor=palette.OrbitPalette.BLUE.value,
    )
    ax[1, 0].set_title("Residual Distribution")
    ax[1, 0].axvline(
        df[residual_col].mean(),
        color=palette.OrbitPalette.ORANGE.value,
        linestyle="--",
        alpha=0.9,
        label="residual mean",
    )
    ax[1, 0].set_ylabel("density")
    ax[1, 0].legend()

    # plot 4 residual qq plot
    if dist == "norm":
        _ = stats.probplot(df[residual_col].values, dist="norm", plot=ax[1, 1])
    elif dist == "t-dist":
        # t-dist qq-plot
        _ = stats.probplot(
            df[residual_col].values, dist=stats.t, sparams=sparams, plot=ax[1, 1]
        )

    # plot 5 residual ACF
    sm.graphics.tsa.plot_acf(
        df[residual_col].values,
        ax=ax[2, 0],
        title="Residual ACF",
        color=palette.OrbitPalette.BLUE.value,
    )
    ax[2, 0].set_xlabel("lag")
    ax[2, 0].set_ylabel("acf")

    # plot 6 residual PACF
    sm.graphics.tsa.plot_pacf(
        df[residual_col].values,
        ax=ax[2, 1],
        title="Residual PACF",
        color=palette.OrbitPalette.BLUE.value,
    )
    ax[2, 1].set_xlabel("lag")
    ax[2, 1].set_ylabel("pacf")
    fig.tight_layout()



import pytest
import numpy as np

from orbit.eda import eda_plot


def test_eda_plot(iclaims_training_data):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])

    # test plotting
    _ = eda_plot.ts_heatmap(
        df=df,
        date_col="week",
        value_col="claims",
        seasonal_interval=52,
        normalization=True,
    )

    var_list = ["trend.unemploy", "trend.filling", "trend.job"]
    _ = eda_plot.correlation_heatmap(df, var_list=var_list)

    _ = eda_plot.dual_axis_ts_plot(
        df=df, var1="trend.unemploy", var2="claims", date_col="week"
    )

    df[["week"] + var_list].melt(id_vars=["week"])
    _ = eda_plot.wrap_plot_ts(df, "week", ["week"] + var_list)



import numpy as np
import pandas as pd
from ..exceptions import IllegalArgument


def get_knot_dates(start_date, knot_idx, freq):
    """
    Parameters
    ----------
    start_date : datetime array
    knot_idx : ndarray
        1D array containing index with `int` type.
    freq : date frequency

    Returns
    -------
    list :
        list of knot dates with provided start date time and indices
    """
    # knot_dates = knot_idx * freq + start_date
    # knot_dates = knot_idx * np.timedelta64(1, freq) + start_date
    dates_lst = pd.date_range(start=start_date, periods=max(knot_idx) + 1, freq=freq)
    knot_dates = dates_lst[knot_idx]

    return knot_dates


def get_dates_delta(start_date, end_date, time_delta):
    """return knot index based on date difference normalized with the number of steps by frequency provided

    Parameters
    ----
    start_date : numpy datetime
    end_date : numpy datetime array
    time_delta : time delta between dates
    """
    date_diff = end_date - start_date
    # can also be deemed as the "knot_idx"
    norm_delta = np.round(date_diff / time_delta).astype(int)

    return norm_delta


def get_knot_idx_by_dist(num_of_obs, knot_distance):
    """function to calculate the knot idx based on num_of_obs and knot_distance."""
    # starts with the the ending point
    # use negative values or simply append 0 to the sequence?
    knot_idx = np.sort(np.arange(num_of_obs - 1, -1, -knot_distance))
    knot_idx = np.round(knot_idx).astype("int")
    if 0 not in knot_idx:
        # knot_idx = np.sort(np.arange(num_of_obs - 1, -1 - knot_distance, -knot_distance))
        knot_idx = np.sort(np.append(knot_idx, 0))

    return knot_idx


def get_knot_idx(
    num_of_obs=None,
    num_of_segments=None,
    knot_distance=None,
    date_array=None,
    knot_dates=None,
):
    """function to calculate and return the knot locations as indices based on
    This function will be used in KTRLite and KTRX model.

    There are three ways to get the knot index:
    1. With number of observations supplied, calculate the knots location and indices based on
    number of segments specified and knot indices will be evenly distributed
    2. With number of observations supplied, calculate the knots location and indices based on
    knot distance specified such that there will be additional knots in the first and end provided
    3. With observations date array and knot dates provided, derive knots location directly based on the
    implied observation frequency provided.

    Parameters
    ----------
    num_of_obs : int
        number of observations to derive segments and knots; will be ignored if knot_dates is not None
    num_of_segments : int
        number of segments, which will be used to calculate the knot distance
    knot_distance : int
        distance between every two knots
    date_array : datetime array
        only used when knot_dates is not None
    knot_dates : list or array of numpy datetime
        list of dates in string format (%Y-%m-%d) or numpy datetime array which will be used as the knot locations

    Returns
    -------
    an array of integers, which are the knot location indices (starts at 0).

    """
    if knot_dates is None and num_of_obs is None:
        raise IllegalArgument("Either knot_dates or num_of_obs needs to be provided.")

    if knot_dates is not None:
        if date_array is None:
            raise IllegalArgument(
                "When knot_dates are supplied, users need to supply date_array as well."
            )
        knot_dates = np.array(knot_dates, dtype="datetime64")

        # filter out
        _knot_dates = pd.to_datetime(
            [
                x
                for x in knot_dates
                if (x <= date_array.max()) and (x >= date_array.min())
            ]
        )

        time_delta = np.diff(date_array).mean()

        knot_idx = get_dates_delta(
            start_date=date_array[0], end_date=_knot_dates, time_delta=time_delta
        )

    elif knot_distance is not None:
        if not isinstance(knot_distance, int):
            raise Exception("knot_distance must be an int.")
        knot_idx = get_knot_idx_by_dist(num_of_obs, knot_distance)

    elif num_of_segments is not None:
        if num_of_segments >= 1:
            knot_distance = (num_of_obs - 1) / num_of_segments
            knot_idx = get_knot_idx_by_dist(num_of_obs, knot_distance)
        else:
            # one single knot at the beginning
            knot_idx = np.array([0])
    else:
        raise Exception(
            "please specify at least one of the followings to determine the knot locations: "
            "knot_dates, knot_distance, or num_of_segments."
        )

    # knot_idx starts with 0; need to add 1 when calculate the fraction
    return knot_idx



import pytest
import pandas as pd
import numpy as np

from orbit.diagnostics.backtest import BackTester
from orbit.diagnostics.metrics import smape
from orbit.models import LGT, DLT
from orbit.diagnostics.plot import (
    plot_predicted_data,
    plot_predicted_components,
    plot_bt_predictions,
    plot_bt_predictions2,
)


@pytest.mark.parametrize(
    "plot_components",
    [["trend", "seasonality", "regression"], ["trend", "seasonality"], ["trend"]],
)
def test_plot_predicted_data(iclaims_training_data, plot_components):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])

    regressor_col = ["trend.unemploy", "trend.filling", "trend.job"]
    test_size = 52
    train_df = df[:-test_size]
    test_df = df[-test_size:]

    lgt = LGT(
        response_col="claims",
        date_col="week",
        regressor_col=regressor_col,
        estimator="stan-map",
        seasonality=52,
        seed=8888,
    )
    lgt.fit(train_df)
    predicted_df = lgt.predict(df=test_df, decompose=True)

    # test plotting
    _ = plot_predicted_data(
        training_actual_df=train_df,
        predicted_df=predicted_df,
        date_col="week",
        actual_col="claims",
        test_actual_df=test_df,
    )

    _ = plot_predicted_components(
        predicted_df=predicted_df, date_col="week", plot_components=plot_components
    )


def test_plot_bt_predictions(iclaims_training_data):
    df = iclaims_training_data
    df["claims"] = np.log(df["claims"])

    regressor_col = ["trend.unemploy", "trend.filling", "trend.job"]

    dlt = DLT(
        date_col="week",
        response_col="claims",
        regressor_col=regressor_col,
        seasonality=52,
        estimator="stan-map",
    )
    bt = BackTester(
        model=dlt, df=df, min_train_len=100, incremental_len=100, forecast_len=20
    )
    bt.fit_predict()
    predicted_df = bt.get_predicted_df()

    # test plotting
    _ = plot_bt_predictions(predicted_df, metrics=smape, ncol=2, include_vline=True)
    _ = plot_bt_predictions2(predicted_df, metrics=smape, include_vline=True)



from ..template.ktrlite import KTRLiteModel
from ..forecaster import MAPForecaster
from ..exceptions import IllegalArgument
from ..estimators.stan_estimator import StanEstimatorMAP
from ..constants.constants import EstimatorsKeys


def KTRLite(
    # level
    level_knot_scale=0.1,
    level_segments=10,
    level_knot_distance=None,
    level_knot_dates=None,
    # seasonality
    seasonality=None,
    seasonality_fs_order=None,
    seasonality_segments=2,
    seasonal_initial_knot_scale=1.0,
    seasonal_knot_scale=0.1,
    degree_of_freedom=30,
    date_freq=None,
    estimator="stan-map",
    suppress_stan_log=True,
    **kwargs,
):
    """
    Parameters
    ----------
    level_knot_scale : float
        sigma for level; default to be .1
    level_segments : int
        the number of segments partitioned by the knots of level (trend)
    level_knot_distance : int
        the distance between every two knots of level (trend)
    level_knot_dates : array like
        list of pre-specified dates for the level knots
    seasonality : int, or list of int
        multiple seasonality
    seasonality_fs_order : int, or list of int
        fourier series order for seasonality
    seasonality_segments : int
        the number of segments partitioned by the knots of seasonality
    seasonal_initial_knot_scale : float
        scale parameter for seasonal regressors initial coefficient knots; default to be 1
    seasonal_knot_scale : float
        scale parameter for seasonal regressors drift of coefficient knots; default to be 0.1.
    degree_of_freedom : int
        degree of freedom for error t-distribution
    date_freq : str
        date frequency; if not supplied, pd.infer_freq will be used to imply the date frequency.
    estimator : string; {'stan-map'}

    Other Parameters
    ----------------
    response_col : str
        Name of response variable column, default 'y'
    date_col : str
        Name of date variable column, default 'ds'
    n_bootstrap_draws : int
        Number of samples to bootstrap in order to generate the prediction interval. For full Bayesian and
        variational inference forecasters, samples are drawn directly from original posteriors. For point-estimated
        posteriors, it will be used to sample noise parameters.  When -1 or None supplied, full Bayesian and
        variational inference forecasters will assume number of draws equal the size of original samples while
        point-estimated posteriors will mute the draw and output prediction without interval.
    prediction_percentiles : list
        List of integers of prediction percentiles that should be returned on prediction. To avoid reporting any
        confident intervals, pass an empty list
    suppress_stan_log : bool
        If False, turn off cmdstanpy logger. Default as False.

    **kwargs:
        additional arguments passed into orbit.estimators.stan_estimator
    """
    _supported_estimators = [EstimatorsKeys.StanMAP.value]

    ktrlite = KTRLiteModel(
        level_knot_scale=level_knot_scale,
        level_segments=level_segments,
        level_knot_distance=level_knot_distance,
        level_knot_dates=level_knot_dates,
        # seasonality
        seasonality=seasonality,
        seasonality_fs_order=seasonality_fs_order,
        seasonality_segments=seasonality_segments,
        seasonal_initial_knot_scale=seasonal_initial_knot_scale,
        seasonal_knot_scale=seasonal_knot_scale,
        degree_of_freedom=degree_of_freedom,
        date_freq=date_freq,
    )
    if estimator == EstimatorsKeys.StanMAP.value:
        ktrlite_forecaster = MAPForecaster(
            model=ktrlite,
            estimator_type=StanEstimatorMAP,
            suppress_stan_log=suppress_stan_log,
            **kwargs,
        )
    else:
        raise IllegalArgument(
            "Invalid estimator. Must be one of {}".format(_supported_estimators)
        )

    return ktrlite_forecaster



import numpy as np
import pandas as pd
from orbit.exceptions import IllegalArgument


def make_fourier_series(n, period, order=3, shift=0):
    """Given time series length, cyclical period and order, return a set of fourier series.

    Parameters
    ----------
    n : int
        Length of time series
    period : int
        Length of a cyclical period. E.g., with daily data, `period = 7` means weekly seasonality.
    order : int
        Number of components for each sin() or cos() series.
    shift : int
        shift of time step/index to generate the series
    Returns
    -------
    2D array-like
        2D array where each column represents the series with a specific order fourier constructed by sin() or cos().
    Notes
    -----
        1. See https://otexts.com/fpp2/complexseasonality.html
        2. Original idea from https://github.com/facebook/prophet under
    """
    t = np.arange(1, n + 1) + shift
    out = list()
    for i in range(1, order + 1):
        x = 2.0 * i * np.pi * t / period
        out.append(np.cos(x))
        out.append(np.sin(x))
    out = np.column_stack(out)
    return out


def make_fourier_series_df(df, period, order=3, prefix="", suffix="", shift=0):
    """Given a data-frame, cyclical period and order, return a set of fourier series in a dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe to supply datetime array for generating fourier series
    period : int
        Length of a cyclical period. E.g., with daily data, `period = 7` means weekly seasonality.
    order : int
        Number of components for each sin() or cos() series.
    prefix : str
        prefix of output columns label
    suffix : str
        suffix of output columns label
    shift : int
        shift of time step/index to generate the series
    Returns
    -------
    out : pd.DataFrame
        data with computed fourier series attached
    fs_cols : list
        list of labels derived from fourier series
    Notes
    -----
        This is calling :func:`make_fourier_series`
    """
    fs = make_fourier_series(df.shape[0], period, order=order, shift=shift)
    fs_cols = []
    for i in range(1, order + 1):
        fs_cols.append("{}fs_cos{}{}".format(prefix, i, suffix))
        fs_cols.append("{}fs_sin{}{}".format(prefix, i, suffix))
    fs_df = pd.DataFrame(fs, columns=fs_cols)
    out = pd.concat([df.reset_index(drop=True), fs_df], axis=1)
    return out, fs_cols


def make_seasonal_dummies(df, date_col, freq, sparse=True, drop_first=True):
    """Based on the frequency input (in pandas.DataFrame style), provide dummies indicator for regression type of
    purpose.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe to supply datetime array for generating series of indicators
    date_col : str
        Label of the date column supply for generating series
    freq : str ['weekday', 'month', 'week']
        Options to pick the right frequency for generating dummies
    sparse : bool
        argument passed into `pd.get_dummies`
    drop_first : bool
        argument passed into `pd.get_dummies`
    Returns
    -------
    out : pd.DataFrame
        data with computed fourier series attached
    fs_cols : list
        list of labels derived from fourier series
    Notes
    -----
        This is calling :func:`pd.get_dummies`
    """
    if freq == "weekday":
        dummies = pd.get_dummies(
            df[date_col].dt.weekday, prefix="wd", sparse=sparse, drop_first=drop_first
        )
    elif freq == "month":
        dummies = pd.get_dummies(
            df[date_col].dt.month, prefix="m", sparse=sparse, drop_first=drop_first
        )
    elif freq == "week":
        dummies = pd.get_dummies(
            df[date_col].dt.week, prefix="w", sparse=sparse, drop_first=drop_first
        )
    else:
        raise IllegalArgument("Invalid argument of freq.")

    cols = dummies.columns.tolist()
    out = pd.concat([df, dummies], axis=1)
    return out, cols


def make_seasonal_regressors(n, periods, orders, labels, shift=0):
    """

    Parameters
    ----------
    n : int
        total length of time steps to generate seasonality; e.g. can simply be the length of your date array or
        dataframe
    periods : list
        list of period (a.k.a seasonality)
    orders : list
        list of fourier series order; needs to be the same length of s
    labels : list
        list of string to label each component
    shift : int
        shift of time step/index to generate the series

    Returns
    -------
    dict :
        a dictionary contains sine-cosine like regressors where keys are mapped by label provided
    """
    out = dict()
    for idx, period in enumerate(periods):
        order = orders[idx]
        label = labels[idx]
        fs = make_fourier_series(n=n, period=period, order=order, shift=shift)
        out[label] = fs
    return out


def moving_average(x, window=1, mode="same"):
    """Compute moving average of a 1-D numpy array

    Parameters
    ----------
    x : array_like
        1D-like array required to compute moving average
    window : int
        window size to compute moving average of the array
    mode : str
        one of the ['full', 'same', valid']. See `numpy.convolve` for details.
    Returns
    -------
        moving average of a 1-D array
    """
    return np.convolve(x, np.ones(window), mode=mode) / window



import numpy as np
from copy import deepcopy
import torch
from enum import Enum

from ..constants.constants import PredictionKeys, TrainingMetaKeys, PredictionMetaKeys
from ..exceptions import IllegalArgument, DataInputException
from .model_template import ModelTemplate
from ..estimators.stan_estimator import StanEstimatorMCMC, StanEstimatorMAP
from ..utils.features import moving_average


# constants for attributes, params and I/Os
class DataInputMapper(Enum):
    """
    mapping from object input to sampler
    """

    # ---------- Seasonality ---------- #
    _SEASONALITY = "SEASONALITY"
    SEASONALITY_SD = "SEASONALITY_SD"
    _SEASONALITY_SM_INPUT = "SEA_SM_INPUT"
    # ---------- Common Local Trend ---------- #
    _LEVEL_SM_INPUT = "LEV_SM_INPUT"
    # handle missing values
    IS_VALID_RESPONSE = "IS_VALID_RES"


class BaseSamplingParameters(Enum):
    """
    base parameters in posteriors sampling
    """

    # ---------- Common Local Trend ---------- #
    LOCAL_TREND_LEVELS = "l"
    LEVEL_SMOOTHING_FACTOR = "lev_sm"
    # ---------- Noise Trend ---------- #
    RESIDUAL_SIGMA = "obs_sigma"


class SeasonalitySamplingParameters(Enum):
    """
    seasonality component related parameters in posteriors sampling
    """

    SEASONALITY_LEVELS = "s"
    SEASONALITY_SMOOTHING_FACTOR = "sea_sm"


class LatentSamplingParameters(Enum):
    """
    latent variables to be sampled
    """

    INITIAL_SEASONALITY = "init_sea"


class ETSModel(ModelTemplate):
    """
    Parameters
    ----------
    seasonality : int
        Length of seasonality
    seasonality_sm_input : float
        float value between [0, 1], applicable only if `seasonality` > 1. A larger value puts
        more weight on the current seasonality.
        If None, the model will estimate this value.
    level_sm_input : float
        float value between [0.0001, 1]. A larger value puts more weight on the current level.
        If None, the model will estimate this value.
    """

    # data labels for sampler
    _data_input_mapper = DataInputMapper
    # used to match name of `*.stan` or `*.pyro` file to look for the model
    _model_name = "ets"
    _supported_estimator_types = [StanEstimatorMAP, StanEstimatorMCMC]

    def __init__(
        self, seasonality=None, seasonality_sm_input=None, level_sm_input=None, **kwargs
    ):
        # estimator is created in base class
        super().__init__(**kwargs)
        self.seasonality = seasonality
        self.seasonality_sd = None

        # fixed smoothing parameters config
        self.seasonality_sm_input = seasonality_sm_input
        self.level_sm_input = level_sm_input

        # set private var to arg value
        # if None set default in _set_default_args()
        self._seasonality = self.seasonality
        self._seasonality_sm_input = self.seasonality_sm_input
        self._level_sm_input = self.level_sm_input
        # handle missing values
        self.is_valid_response = None

        self._set_static_attributes()
        self._set_model_param_names()

    def _set_static_attributes(self):
        """Set attributes which are independent from data"""
        # setting defaults and proper data type
        if self.seasonality_sm_input is None:
            self._seasonality_sm_input = -1
        if self.level_sm_input is None:
            self._level_sm_input = -1
        elif self.level_sm_input < 0 or self.level_sm_input > 1:
            raise IllegalArgument(
                "only values between [0, 1] are supported for level_sm_input "
                "to build a model with meaningful trend."
            )
        if self.seasonality is None:
            self._seasonality = -1

    def set_dynamic_attributes(self, df, training_meta):
        """Set attributes which are dependent on data"""
        # compute data-driven prior for seasonality
        if self._seasonality > 1:
            response = training_meta[TrainingMetaKeys.RESPONSE.value]
            response_ma = moving_average(
                response, window=self._seasonality, mode="same"
            )
            adjusted_response = response - response_ma
            # to estimate the "across-group" s.d. as a seasonality prior
            ss = np.zeros(self._seasonality)
            for idx in range(self._seasonality):
                ss[idx] = np.nanmean(adjusted_response[idx :: self._seasonality])
            self.seasonality_sd = np.nanstd(ss)
        else:
            # should not be used anyway; just a placeholder
            self.seasonality_sd = training_meta[TrainingMetaKeys.RESPONSE_SD.value]

        self._set_valid_response(training_meta)

    def _set_valid_response(self, training_meta):
        response = training_meta[TrainingMetaKeys.RESPONSE.value]
        self.is_valid_response = (~np.isnan(response)).astype(int)
        # raise exception if the first response value is missing
        if self.is_valid_response[0] == 0:
            raise DataInputException(
                "The first value of response column {} cannot be missing..".format(
                    training_meta[TrainingMetaKeys.RESPONSE_COL.value]
                )
            )

    def set_init_values(self):
        """Override function from base class"""
        init_values = dict()
        if self._seasonality > 1:
            init_sea = np.clip(
                np.random.normal(loc=0, scale=0.05, size=self._seasonality - 1),
                -1.0,
                1.0,
            )
            init_values[LatentSamplingParameters.INITIAL_SEASONALITY.value] = init_sea
            self._init_values = init_values

    def _set_model_param_names(self):
        """Set posteriors keys to extract from sampling/optimization api"""
        self._model_param_names += [param.value for param in BaseSamplingParameters]

        # append seasonality param names
        if self._seasonality > 1:
            self._model_param_names += [
                param.value for param in SeasonalitySamplingParameters
            ]

    def predict(
        self,
        posterior_estimates,
        df,
        training_meta,
        prediction_meta,
        include_error=False,
        **kwargs,
    ):
        """Vectorized version of prediction math"""
        ################################################################
        # Prediction Attributes
        ################################################################
        # n_forecast_steps = prediction_meta[PredictionMetaKeys.FUTURE_STEPS.value]
        start = prediction_meta[PredictionMetaKeys.START_INDEX.value]
        trained_len = training_meta[TrainingMetaKeys.NUM_OF_OBS.value]
        full_len = prediction_meta[PredictionMetaKeys.END_INDEX.value]

        ################################################################
        # Model Attributes
        ################################################################
        model = deepcopy(posterior_estimates)
        for k, v in model.items():
            model[k] = torch.from_numpy(v)

        # We can pull any arbitrary value from teh dictionary because we hold the
        # safe assumption: the length of the first dimension is always the number of samples
        # thus can be safely used to determine `num_sample`. If predict_method is anything
        # other than full, the value here should be 1
        arbitrary_posterior_value = list(model.values())[0]
        num_sample = arbitrary_posterior_value.shape[0]

        # seasonality components
        seasonality_levels = model.get(
            SeasonalitySamplingParameters.SEASONALITY_LEVELS.value
        )
        seasonality_smoothing_factor = model.get(
            SeasonalitySamplingParameters.SEASONALITY_SMOOTHING_FACTOR.value
        )

        # trend components
        level_smoothing_factor = model.get(
            BaseSamplingParameters.LEVEL_SMOOTHING_FACTOR.value
        )
        local_trend_levels = model.get(BaseSamplingParameters.LOCAL_TREND_LEVELS.value)
        residual_sigma = model.get(BaseSamplingParameters.RESIDUAL_SIGMA.value)

        ################################################################
        # Seasonality Component
        ################################################################

        # calculate seasonality component
        if self._seasonality > 1:
            if full_len <= seasonality_levels.shape[1]:
                seasonal_component = seasonality_levels[:, :full_len]
            else:
                seasonality_forecast_length = full_len - seasonality_levels.shape[1]
                seasonality_forecast_matrix = torch.zeros(
                    (num_sample, seasonality_forecast_length), dtype=torch.double
                )
                seasonal_component = torch.cat(
                    (seasonality_levels, seasonality_forecast_matrix), dim=1
                )
        else:
            seasonal_component = torch.zeros((num_sample, full_len), dtype=torch.double)

        ################################################################
        # Trend Component
        ################################################################

        # calculate level component.
        # However, if predicted end of period > training period, update with out-of-samples forecast
        if full_len <= trained_len:
            trend_component = local_trend_levels[:, :full_len]
            # in-sample error are iids
            if include_error:
                error_value = np.random.normal(
                    loc=0,
                    scale=residual_sigma.unsqueeze(-1),
                    size=trend_component.shape,
                )

                error_value = torch.from_numpy(error_value).double()
                trend_component += error_value
        else:
            trend_component = local_trend_levels
            trend_forecast_length = full_len - trained_len
            trend_forecast_init = torch.zeros(
                (num_sample, trend_forecast_length), dtype=torch.double
            )
            # in-sample error are iids
            if include_error:
                error_value = np.random.normal(
                    loc=0,
                    scale=residual_sigma.unsqueeze(-1),
                    size=trend_component.shape,
                )

                error_value = torch.from_numpy(error_value).double()
                trend_component += error_value

            trend_component = torch.cat((trend_component, trend_forecast_init), dim=1)

            last_local_trend_level = local_trend_levels[:, -1]

            for idx in range(trained_len, full_len):
                trend_component[:, idx] = last_local_trend_level

                if include_error:
                    error_value = np.random.normal(
                        scale=residual_sigma, size=num_sample
                    )
                    error_value = torch.from_numpy(error_value).double()
                    trend_component[:, idx] += error_value

                new_local_trend_level = (
                    level_smoothing_factor * trend_component[:, idx]
                    + (1 - level_smoothing_factor) * last_local_trend_level
                )

                if self._seasonality > 1 and idx + self._seasonality < full_len:
                    seasonal_component[:, idx + self._seasonality] = (
                        seasonality_smoothing_factor.flatten()
                        * (
                            trend_component[:, idx]
                            + seasonal_component[:, idx]
                            - new_local_trend_level
                        )
                        + (1 - seasonality_smoothing_factor.flatten())
                        * seasonal_component[:, idx]
                    )

                last_local_trend_level = new_local_trend_level

        ################################################################
        # Combine Components
        ################################################################

        # trim component with right start index
        trend_component = trend_component[:, start:]
        seasonal_component = seasonal_component[:, start:]

        # sum components
        pred_array = trend_component + seasonal_component

        pred_array = pred_array.numpy()
        trend_component = trend_component.numpy()
        seasonal_component = seasonal_component.numpy()

        out = {
            PredictionKeys.PREDICTION.value: pred_array,
            PredictionKeys.TREND.value: trend_component,
            PredictionKeys.SEASONALITY.value: seasonal_component,
        }

        return out
